from ast import If
import numpy as np
from scipy import signal
from scipy import stats
from .tools import PSDgen, random_gaussian
import warnings

class SpectralData(object):
    """
    SpectralData object contains data, based on input signal time-history:
    power spectral density (PSD), spectral moments, spectral band estimators
    and others parameters, required for subsequent fatigue-life estimation [1].

    Notes
    -----
    Fatigue-life estimators inherit from instance of class SpectralData, so a 
    primitive caching is used. With class instantiation, values that are expected
    to be used subseqeuntly are calculated. The properties are created and then 
    called. Calculation is done only at the beginning and only once for following
    spectral method:

    - :class:`~FLife.freq_domain.Narrowband`
    - :class:`~FLife.freq_domain.WirschingLight`
    - :class:`~FLife.freq_domain.OrtizChen`
    - :class:`~FLife.freq_domain.Alpha075`
    - :class:`~FLife.freq_domain.TovoBenasciutti`
    - :class:`~FLife.freq_domain.Dirlik`
    - :class:`~FLife.freq_domain.ZhaoBaker`
    - :class:`~FLife.freq_domain.Park`
    - :class:`~FLife.freq_domain.JunPark`

    For multimodal fatigue-life estimators, spectral data are calculated a 
    second time, referring to each mode independently: 

    - :class:`~FLife.freq_domain.JiaoMoan`
    - :class:`~FLife.freq_domain.SakaiOkamura`
    - :class:`~FLife.freq_domain.FuCebon`
    - :class:`~FLife.freq_domain.ModifiedFuCebon`
    - :class:`~FLife.freq_domain.Low`
    - :class:`~FLife.freq_domain.Low2014`
    - :class:`~FLife.freq_domain.GaoMoan`
    - :class:`~FLife.freq_domain.Lotsberg`
    - :class:`~FLife.freq_domain.HuangMoan`
    - :class:`~FLife.freq_domain.SingleMoment`
    - :class:`~FLife.freq_domain.BandsMethod`

    SpectralData is also taken as base for Rainflow class, the fatigue-life 
    reference:

    - :class:`~FLife.time_domain.Rainflow`

    References
    ----------
    [1] Janko Slavič, Matjaž Mršnik, Martin Česnik, Jaka Javh, Miha Boltežar. 
        Vibration Fatigue by Spectral Methods, From Structural Dynamics to Fatigue Damage
        – Theory and Experiments, ISBN: 9780128221907, Elsevier, 1st September 2020

    Example
    -------
    Import modules, define time- and frequency-domain data
    
    >>> import FLife
    >>> import pyExSi as es
    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> # time-domain data
    >>> N = 2 ** 16  # number of data points of time signal
    >>> fs = 2048  # sampling frequency [Hz]
    >>> t = np.arange(0, N) / fs  # time vector
    >>> # frequency-domain data
    >>> M = N // 2 + 1  # number of data points of frequency vector
    >>> freq = np.arange(0, M, 1) * fs / N  # frequency vector
    >>> PSD_lower = es.get_psd(freq, 20, 60, variance = 5)  # lower mode of random process
    >>> PSD_higher = es.get_psd(freq, 100, 120, variance = 2)  # higher mode of random process
    >>> PSD = PSD_lower + PSD_higher # bimodal one-sided flat-shaped PSD

    Get Gaussian stationary signal, instantiate SpectralData object and plot PSD

    >>> rg = np.random.default_rng(123) # random generator seed
    >>> x = es.random_gaussian(N, PSD, fs, rg) # Gaussian stationary signal
    >>> sd = FLife.SpectralData(input=x, dt=1/fs) # SpectralData instance
    >>> plt.plot(sd.psd[:,0], sd.psd[:,1]) 
    >>> plt.xlabel('Frequency [Hz]')
    >>> plt.ylabel('PSD')
    >>> print(f'Irregularity factor: {sd.alpha2:.4f}') # irregularity factor

    Calculate spectral width parameter (epsilon = (1 - alpha2**2)**.5), with
    alpha2 obtained using function :meth:`.SpectralData.get_bandwidth_estimator`

    >>> # one element is in returned array, as entire PSD is taken as one band
    >>> alpha2, = sd.get_bandwidth_estimator(sd.PSD_splitting, i=2)
    >>> epsilon = np.sqrt(1 - alpha2**2)
    >>> print(f'Spectral width parameter: {epsilon:.4f}')

    Define S-N curve parameters and get fatigue-life estimatate

    >>> C = 1.8e+22  # S-N curve intercept [MPa**k]
    >>> k = 7.3 # S-N curve inverse slope [/]
    >>> dirlik = FLife.Dirlik(sd) # Dirlik's fatigue-life estimator
    >>> tb = FLife.TovoBenasciutti(sd) #Tovo-Benasciutti's fatigue-life estimator
    >>> jm = FLife.JiaoMoan(sd, PSD_splitting=('userDefinedBands', [80,150]))  # Jiao-Moan's fatigue-life estimator
    >>> print(f'Dirlik: {dirlik.get_life(C,k):.3e} s.')   
    >>> print(f'Tovo-Benasciutti, method 1: {tb.get_life(C,k,method='method 1'):.3e} s.')
    >>> print(f'Jiao-Moan: {jm.get_life(C,k):.3e} s.')   
  
    Plot segmented PSD, used in Jiao-Moan method

    >>> lower_band_index, upper_band_index= jm.band_stop_indexes
    >>> plt.plot(sd.psd[:,0], sd.psd[:,1])
    >>> plt.vlines(sd.psd[:,0][lower_band_index], 0, np.max(sd.psd[:,1]), 'k', linestyles='dashed', alpha=.5)
    >>> # lower band
    >>> plt.fill_between(sd.psd[:lower_band_index,0], sd.psd[:lower_band_index,1], 'o', label='lower band', alpha=.2, color='blue')
    >>> plt.vlines(sd.psd[:,0][upper_band_index], 0, np.max(sd.psd[:,1]), 'k', linestyles='dashed', alpha=.5)
    >>> # upper band
    >>> plt.fill_between(sd.psd[lower_band_index:upper_band_index,0], sd.psd[lower_band_index:upper_band_index,1], 'o', label='upper band', alpha=.5, color ='orange')
    >>> plt.xlabel('Frequency [Hz]')
    >>> plt.ylabel('PSD')
    >>> plt.xlim(0,300)
    >>> plt.legend()

    Calculate irregularity factor for both PSD bands, defined in Jiao-Moan method

    >>> alpha2_lower_band, alpha2_higher_band = jm.spectral_data.get_bandwidth_estimator(jm.PSD_splitting, i=2) #or sd.get_bandwidth_estimator(jm.PSD_splitting, i=2)
    >>> print(f'Irregularity factor, lower band: {alpha2_lower_band:.4f}')
    >>> print(f'Irregularity factor, higher band:{alpha2_higher_band:.4f}')
    """

    def __init__(self, input=None, window='hann', nperseg=1280,
                 noverlap=None, psd_trim_length=None, 
                 T=None, fs=None, rg=None,
                 **kwargs):

        """Call parent constructor, analyse input and define fatigue life parameters and constants.
        
        :param input: str, tuple
                Defaults to 'GUI', where PSD is provided via GUI. Another option is dictionary,
                where input can be provided as psd with keys: 'PSD'(PSD array) and 'f'(frequency vector array),
                or as time history with keys: 'time_history' and 'dt', or as amplitude spectrum with keys: 'amplitude_spectrum' and 'f'.
                
                For uniaxial PSD input, PSD array should be the shape of (f), 
                for multiaxial PSD input, PSD array should be a matrix with the shape of (f,6,6) for 3D stress state or (f,3,3) for 2D stress state.
                For multi-point PSD input (whole FEM model), PSD array should be a matrix
                with the shape of (N,f,6,6) for 3D stress state or (N,f,3,3) for 2D stress state, 
                where N is the number of points on the model.
                
                For uniaxial amplitude spectrum input, amplitude_spectrum array should be the shape of (f), 
                for multiaxial amplitude spectrum input, amplitude_spectrum array should be a matrix
                with the shape of (f,6) for 3D stress state or (f,3) for 2D stress state.
                For multi-point amplitude spectrum input (whole FEM model), amplitude_spectrum array should be a matrix
                with the shape of (N,f,6) for 3D stress state or (N,f,3) for 2D stress state, 

        :param window: str or tuple or array_like, optional
                Desired window to use. Defaults to ‘hann’.
        :param nperseg:  int, optional
                Length of each segment. Defaults to 1280.
        :param noverlap: int, optional
                Number of points to overlap between segments. If None, 
                noverlap = nperseg / 2. Defaults to None.
        :param psd_trim_length: int, optional
                Number of frequency points to be used for PSD.
                Defaults to None.
        :param T: int, float, optional
                Length of time history when random process is defined by `input` parameter 'GUI'
                or by PSD or by amplitude spectrum. If T and fs are provided, time histoy is generated.
                Defaults to None.
        :param fs: int, float, optional
                Sampling frequency of time history when random process is defined by `input` parameter 
                'GUI' or by PSD or by amplitude spectrum. If T and fs are provided, time histoy is generated.
                Defaults to None.
        :param rg: numpy.random._generator.Generator, optional
                Random generator controls phase of generated time history, when `input` is 'GUI' or 
                (PSD, frequency vector).
                Defaults to None.

        Attributes
        ----------
        data: array 1D
                Signal values with time spacing dt
        dt: float
                Time between discreete signal values    
        moments: array 1D
                Spectral moments from m0 to m4    
        psd: array 2D
                Normalized power spectral density
        t: float
                Length of signal in seconds, N * dt
        PSD_splitting: tuple
                PSD_splitting[0] is PSD spliting method, PSD_splitting[1] is method argument. 
                Splitting methods:

                - 'userDefinedBands', PSD_splitting[1] must be of type list or tupple, with N 
                  elements specifying upper band frequencies of N random processes.
                - 'equalAreaBands', PSD_splitting[1] must be of type int, specifying N random processes.

                Defaults to ('equalAreaBands', 1).

        Raises
        ------
        ValueError, for unknown input.
        """
        if 'dt' in kwargs:
            warnings.warn('Parameter `dt` has been deprecated since version 1.2 and will be removed in the future.' + 
            '\nSampling interval of time signal should be given as second element of tuple `input`.')
            input = (input, kwargs['dt'])
        
        # Default input is by GUI
        if input == None or input == 'GUI':
            self.psd = PSDgen().get_PSD()
            
            # needed parameters for time-history generation
            if T is not None and fs is not None:
                self._set_time_history(f=self.psd[:,0], psd=self.psd[:,1], T=T, fs=fs, **kwargs)


        # If input is tuple, dictionary is created (backwards compatibility)
        elif isinstance(input, tuple) and len(input) == 2:
            warnings.warn('Tuple input has been deprecated. Use dictionary input instead.')
            # If input is time signal
            if isinstance(input[0], np.ndarray) and isinstance(input[1], (int, float)):
                input = {'time_history': input[0], 'dt': input[1]}
            # or PSD
            elif isinstance(input[0], np.ndarray) and isinstance(input[1], np.ndarray):
                input = {'PSD': input[0], 'f': input[1]}

            elif isinstance(input[0], str) and isinstance(input[1], (int, float)):
                    warnings.warn('Path to file option has been deprecated since version 1.2 and will be removed in the future.')
                    self.data = self._readf(input[0])
                    self.dt = input[1] # Sampling interval
                    self.t = self.dt * self.data.size # length of signal time-history
                    self.psd = self._calculate_psd(self.data, fs=1.0/self.dt, window=window,
                                                nperseg=nperseg, noverlap=noverlap,
                                                trim=psd_trim_length)
                                                
                    

        # Input is dictionary a) PSD and frequency vector or b) time history and sampling interval
        if isinstance(input, dict):
            if 'time_history' in input and 'dt' in input:
                self.data = input['time_history']
                self.dt = input['dt']
                self.t = self.dt * self.data.size
                self.psd = self._calculate_psd(self.data, fs=1.0/self.dt, window=window,
                                            nperseg=nperseg, noverlap=noverlap,
                                            trim=psd_trim_length)
            elif 'PSD' in input and 'f' in input:
                #multiaxial PSD
                if input['PSD'].ndim>1:
                    self.multiaxial_psd = (input['PSD'],input['f'])

                    # Check dimension of multiaxial stress PSD
                    if self.multiaxial_psd[0].ndim == 3 or self.multiaxial_psd[0].ndim == 4 and (
                        (self.multiaxial_psd[0].shape[-1] == 6 and self.multiaxial_psd[0].shape[-2] == 6) or
                        (self.multiaxial_psd[0].shape[-1] == 3 and self.multiaxial_psd[0].shape[-2] == 3)
                    ):     
                        if self.multiaxial_psd[0].ndim == 4:
                            self.multipoint = True

                        if T is not None and fs is not None:
                            self.t = T
                            self.fs = fs
                    else:
                        raise Exception('Input Error. PSD matrix should be the size of (f,6,6) for 3D stress state or (f,3,3) for 2D stress state')

                    if T is not None and fs is not None:
                        self._check_fs(f=input['f'], psd=input['PSD'], T=T, fs=fs, multipoint=True, **kwargs)
            
                # Uniaxial PSD
                elif input['PSD'].ndim==1:
                    psd = input['PSD']
                    f = input['f']
                    self.psd = np.column_stack((f, psd))
                    # needed parameters for time-history generation
                    if T is not None and fs is not None:
                        self._set_time_history(f=f, psd=psd, T=T, fs=fs, **kwargs)
            
            elif 'amplitude_spectrum' in input and 'f' in input:
                #multiaxial amplitude spectrum
                if input['amplitude_spectrum'].ndim>1:
                    self.multiaxial_amplitude_spectrum = (input['amplitude_spectrum'],input['f'])

                    # Check dimension of multiaxial amplitude spectrum
                    if self.multiaxial_amplitude_spectrum[0].ndim == 2 or self.multiaxial_amplitude_spectrum[0].ndim == 3 and (
                        (self.multiaxial_amplitude_spectrum[0].shape[-1] == 6) or
                        (self.multiaxial_amplitude_spectrum[0].shape[-1] == 3)
                    ):     
                        if self.multiaxial_amplitude_spectrum[0].ndim == 3:
                            self.multipoint = True

                        if T is not None and fs is not None:
                            self.t = T
                            self.fs = fs
                    else:
                        raise Exception('Input Error. PSD matrix should be the size of (f,6,6) for 3D stress state or (f,3,3) for 2D stress state')

                    if T is not None and fs is not None:
                        self._check_fs(f=input['f'], psd=input['amplitude_spectrum'], T=T, fs=fs, multipoint=True, **kwargs)
                
                # Uniaxial amplitude spectrum
                elif input['amplitude_spectrum'].ndim==1:
                    print('Input amplitude spectrum is uniaxial')
                    amplitde_spectrum = input['amplitude_spectrum']
                    f = input['f']
                    psd = amplitde_spectrum**2/(f[1]-f[0])
                    self.psd = np.column_stack((f, psd))
                    # needed parameters for time-history generation
                    if T is not None and fs is not None:
                        self._set_time_history(f=f, psd=psd, T=T, fs=fs, **kwargs)

            else:
                raise Exception('Unrecognized Input Error. `input` should be dictionary with keys `PSD` and `f` or `time_history` and `dt`.')

        else:
            raise Exception('Unrecognized Input Error') 

        if hasattr(self,'psd'):
            self.PSD_splitting = ('equalAreaBands', 1) 
            self._calculate_coefficients()

    def _readf(self, filename):
        """Read input file and extract values in form of array (float).
        
        Values inside txt file must be separated with whitespace, e.g. space or newline.
        
        Parameters
        ----------
        filename : string
                   path to file
        
        Returns
        -------
        1D array filled with values. 
        """
        f = open(filename)
        data = np.array(map(float,f.read().split()))
        f.close()
    
        return data
    
    def _check_fs(self, f, psd, T=None, fs=None, multipoint=False, **kwargs):
        
        #if input is multi point, multiaxial
        if multipoint:
            f_max_indx = np.where(psd>0)[1][-1]
        else:
            f_max_indx = np.where(psd>0)[0][-1]
        f_max = f[f_max_indx]
                
        # Sampling frequency check
        if fs < 2*f_max:
            raise Exception('Parameter `fs` should be higher. It should be fs >= 2*fmax.')
                    
        if fs < 10*f_max:
            warnings.warn(f'Parameter `fs` should be higher. It is suggested to use fs>=10*fmax.')

    def _set_time_history(self, f, psd, T=None, fs=None, **kwargs):
        """Generates and set time-history of signal on basis of PSD and frequency vector.
        """   
        if T is not None and fs is not None:
            
            self._check_fs(f, psd, T, fs, multipoint=False, **kwargs)
            # get time history
            time, signal = random_gaussian(freq=f, PSD=psd, T=T, fs=fs, **kwargs) # stationary, normally distributed

            #set data
            self.data = signal
            self.dt = 1.0/fs  # Sampling interval
            self.t = T # signal time-history duration
            
        else:
            warnings.warn('Parameters `T` and `fs` should both be specified. Time history is not set.')

    def _calculate_psd(self, data, fs=1, window='hamming', nperseg=10280, noverlap=None, trim=None):
        """Calculates PSD using welch estimator.
        """
        f, p = signal.welch(data, fs=fs, window=window,
                        nperseg=nperseg, noverlap=noverlap)

        psd = np.vstack((f, p)).transpose()

        if trim is None:
            return psd
        else:
            df = fs / nperseg
            trim_idx = int(np.floor(trim * df))

            return psd[:trim_idx, :trim_idx]

    def _calculate_coefficients(self):
        """Calculates coefficients, needed in spectral-fatigue-life estimates. Coefficients are 
        calculated with regard to entire PSD.
        """
        self.moments = np.ravel(self.get_spectral_moments(self.PSD_splitting, moments=[0,1,2,3,4]))
        self.nu, = self.get_nup(self.PSD_splitting)
        self.m_p, = self.get_mp(self.PSD_splitting)
        self.alpha075, = self.get_bandwidth_estimator(self.PSD_splitting, i=0.75)
        self.alpha1, = self.get_bandwidth_estimator(self.PSD_splitting, i=1)
        self.alpha2, = self.get_bandwidth_estimator(self.PSD_splitting, i=2)

    def _get_spectral_moment(self, psd, i):
        """Calculates i-th spectral moment. 
        """
        if not isinstance(i, (int, float)):
            raise TypeError('Parameter `i` must be of type int or float')

        f = psd[:, 0]
        p = psd[:, 1]
        
        if np.__version__>='2.0.0':
            trapezoid = np.trapezoid
        else:
            trapezoid = np.trapz
        
        return trapezoid((2*np.pi*f)**i * p, f)


    def get_spectral_moments(self, PSD_splitting=None, moments=[0,1,2,3,4]):
        """Returns spectral moments, specified by `moments`.
        Depending on parameter `PSD_splitting`, function returns
        an array of shape (M x N), where M is the number of PSD segments
        and N is length of parameter `moments`. For SpectralData, M defaults
        to 1.

        :param PSD_splitting: :attr:`.PSD_splitting`
        :param moments: list
            List of spectral momemnts to be calculated. Its elements must be of type
            int or float.
        :return: Spectral moments.
        :rtype: numpy.ndarray; 
            An array object of shape (M x N).
        """
        if PSD_splitting == None:
            PSD_splitting = self.PSD_splitting 
        band_stop_indexes = self._get_band_stop_frequency(PSD_splitting)
        m_list = list()

        for index in range(len(band_stop_indexes)):
            if index == 0:
                m_list.append([self._get_spectral_moment(self.psd[:band_stop_indexes[index]+1,:], i) for i in moments])
            else:
                m_list.append([self._get_spectral_moment(self.psd[band_stop_indexes[index-1]:band_stop_indexes[index]+1,:], i) for i in moments])

        return np.array(m_list)

    def get_bandwidth_estimator(self, PSD_splitting=None, i=2): 
        """Calculates bandwidth estimator alpha_i [1]. Takes parameter `PSD_splitting`
        for reference to PSD segmentation.

        :param PSD_splitting: :attr:`.PSD_splitting`
        :param i: [int,float]
            Bandwith estimator index; alpha_i = m_i/((m0*m_2i)**.5), where m_i represents
            i-th spectral moment.
        :return: bandwidth estimator;
            An array object of length N, containing bandwidth estimator for N bands.
        :rtype: numpy.ndarray
        """
        if PSD_splitting == None:
            PSD_splitting = self.PSD_splitting 

        if not isinstance(i, (int, float)):
            raise TypeError('Parameter `i` must be of type int or float')
        
        moments = self.get_spectral_moments(PSD_splitting, moments=[0,i,2*i])
        alpha_list = list()
        for band_moments in moments:
            alpha_i  = band_moments[1] / np.sqrt(band_moments[0] * band_moments[2])
            alpha_list.append(alpha_i)

        return np.array(alpha_list)

    def get_vanmarcke_parameter(self, PSD_splitting=None):
        """Calculates Vanmarcke bandwidth parameter epsilon_V[1]. Takes parameter 
        `PSD_splitting` for reference to PSD segmentation.

        :param PSD_splitting: :attr:`.PSD_splitting`
        :return: bandwidth estimator;
            An array object of length N, containing vanmarcke's parameter for N bands.
        :rtype: numpy.ndarray
        """
        if PSD_splitting == None:
            PSD_splitting = self.PSD_splitting 

        alpha = self.get_bandwidth_estimator(PSD_splitting, i=1)
        epsV_list = list()
        for alpha_1 in alpha:
            epsV_i  = np.sqrt(1 - alpha_1**2)
            epsV_list.append(epsV_i)

        return np.array(epsV_list)

    def get_nup(self, PSD_splitting=None):
        """Calculates nu_p; expected frequency of positive slope zero crossing [1].
        Takes parameter `PSD_splitting` for reference to PSD segmentation.

        :param PSD_splitting: :attr:`.PSD_splitting`
        :return: bandwidth estimator;
            An array object of length N, containing nup for N bands.
        :rtype: numpy.ndarray
        """
        moments = self.get_spectral_moments(PSD_splitting, moments=[0,2])
        nup_list = list()
        for band_moments in moments:
            nup_i  = 1/(2*np.pi) * np.sqrt(band_moments[1] / band_moments[0])
            nup_list.append(nup_i)

        return np.array(nup_list)

    def get_mp(self, PSD_splitting=None):
        """Calculates m_p; expected peak frequency [1]. Takes parameter 
        `PSD_splitting` for reference to PSD segmentation.

        :param PSD_splitting: :attr:`.PSD_splitting`
        :return: bandwidth estimator;
            An array object of length N, containing m_p for N bands.
        :rtype: numpy.ndarray
        """
        moments = self.get_spectral_moments(PSD_splitting, moments=[2,4])
        mp_list = list()
        for band_moments in moments:
            mp_i  = 1/(2*np.pi) * np.sqrt(band_moments[1] / band_moments[0])
            mp_list.append(mp_i)

        return np.array(mp_list)

    def get_peak_PDF(self, s):
        """Returns peak amplitude PDF(Probability Density Function) as a function of stress s [1].

        :param s:  numpy.ndarray
            Stress vector.
        :return: function pdf(s)
        """
        m0 = self.moments[0]
        alpha2 = self.alpha2

        def pdf(s):
            px = np.sqrt(1.0 - alpha2**2)/np.sqrt(2.0 * np.pi * m0) * \
                np.exp( - (s**2) / (2.0 * m0 * (1.0 - alpha2**2))) +\
                alpha2*s/m0 * np.exp( - (s**2) / (2*m0)) * \
                stats.norm.cdf((alpha2 * s) / (np.sqrt(m0 * (1 - alpha2**2))))
            return px
        return pdf(s)

    def _get_band_stop_frequency(self, PSD_splitting=None):
        """Returns stop band frequency indexes of segmentated PSD. Takes parameter 
        `PSD_splitting` for reference to PSD segmentation.
        
        :param PSD_splitting: tuple;
            splitting[0] is PSD spliting method, splitting[1] is method argument. 
            Splitting methods:
            - `userDefinedBands`, PSD_splitting[1] must be of type list or tupple,
              with N elements specifying upper band frequencies of N random processes.
            - `equalAreaBands`, PSD_splitting[1] must be of type int, specifying N
              random processes.
        :return freq_indx: tuple
            Upper band frequency indexes of segmentated PSD.
        """
        if PSD_splitting == None:
            PSD_splitting = self.PSD_splitting 

        method_dict = {'equalAreaBands': self._equalAreaBands,
                        'userDefinedBands': self._userDefinedBands
                    }

        if isinstance(PSD_splitting, tuple):
            method = PSD_splitting[0]
            arg = PSD_splitting[1]
            try:
                freq_indx = method_dict[method](arg)
            except:
                raise ValueError(f'Unknown splitting method: {method:s}')
        else:
            raise TypeError('Parameter `PSD_splitting` must be of type tuple.') 
    
        return freq_indx

    def _userDefinedBands(self, freq_list):
        #data type checking
        if isinstance(freq_list, (list, tuple)):
            pass
        else:
            raise TypeError('Parameter `PSD_splitting[1]` must be of type list or tuple.')

        for freq in freq_list:
            if isinstance(freq, (int,float)):
                pass
            else:
                raise TypeError('Parameter `PSD_splitting[1]` elements must be of type int or float.')

        psd_freq = self.psd[:,0]
        freq_indx = list()
        for n in range(len(freq_list)):
            freq_indx.append(np.abs(psd_freq - freq_list[n]).argmin())

        return tuple(freq_indx)

    def _equalAreaBands(self, N):
        if isinstance(N, int):
            pass
        else:
            raise TypeError('Parameter `PSD_splitting[1]` must be of type int.')

        psd_val = self.psd[:,1]
        q_area = np.sum(psd_val) # (not the REAL area because x-axis unit is omitted)
        
        # -- Calculate a cumulative sum vector of psd
        c_psd = np.cumsum(psd_val)
        
        # -- Find where cumulative sum is equal to 1/N, 2/N, ...,  N/N of Area
        freq_indx = list()
        for n in range(N):
            freq_indx.append(np.abs(c_psd - (n+1) * q_area/N).argmin())

        return tuple(freq_indx)


import pyvista as pv
import numpy as np
from matplotlib.figure import Figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas


def set_mesh(self,filename):
    """
    Set the mesh to be visualized. The mesh is loaded from a file. Supports file formats accepted by PyVista https://pyvista.org/.
    :param filename: str
            The filename of the mesh to be visualized and used with pick_point method.
    """
    self.mesh = pv.read(filename)


def pick_point(self):
    """
    Visualize and interact with a mesh using PyVista and Matplotlib.
    This method allows users to pick points on the mesh and
    dynamically update the displayed data using a slider.
    Select the point by right clicking on the mesh. Use left click to rotate the mesh and middle to pan.
    When the point is selected, the equivalent stress PSD data at that point is written to self.psd argument to be used in fatigue analysis.
    """
    if not hasattr(self, 'mesh'):
        raise AttributeError("Mesh not set. Use set_mesh method to set the mesh.")
    else:
        # Create the plotter object
        p = pv.Plotter()
        f = Figure()
        canvas = FigureCanvas(f)
        ax = f.add_subplot(111)
        f.tight_layout(pad=2)

        def refresh_psd_graph():
            ax.clear()
            if hasattr(self, 'psd'):
                ax.semilogy(self.psd[:, 0], self.psd[:, 1])
                ax.axvline(self.psd[freq_index, 0], color='C1')
                ax.set_xlabel('frequency [Hz]')
                ax.set_ylabel('PSD [Unit^2/Hz]')
            f.tight_layout()

            # Redraw the Matplotlib figure
            
            f.canvas.draw()

        # Initial scalar index (start in the middle of the frequency range)
        freq_index = len(self.eq_psd_multipoint[1]) // 2
        self.mesh['eq_stress'] = self.eq_psd_multipoint[0][:, freq_index]  # Initial scalars

        # Add your mesh and keep a reference to the actor

        actor = p.add_mesh(
            self.mesh,
            scalars='eq_stress',
            pickable=True,
            show_scalar_bar=True,
            scalar_bar_args={
                
                'position_x': 0.5,
                'position_y': 0.1,
                'width': 0.45
            }
        )

        custom_camera = pv.Camera()
        
        # Set the custom camera
        p.camera = custom_camera

        # Define a callback function for the point picker
        def point_picker_callback(point):
            nonlocal freq_index  # Ensure freq_index is accessible
            if point is not None:
                # Find the index of the picked point in the mesh
                point_id = self.mesh.find_closest_point(point)
                self.point_id = point_id
                # Update the eq_stress data with the PSD data at the picked point
                self.set_eq_stress(self.eq_psd_multipoint[0][point_id], self.eq_psd_multipoint[1])
                refresh_psd_graph()
                p.render()

        # Define a callback function for the slider
        def slider_callback(value):
            nonlocal freq_index  # Access the freq_index from the outer scope
            freq_index = int(round(value))  # Round the slider value to the nearest whole number
            # Directly modify the scalars of the mesh in-place
            self.mesh['eq_stress'] = self.eq_psd_multipoint[0][:, freq_index]
            refresh_psd_graph()
            vmax = 0.8*np.max(self.mesh['eq_stress'])
            # Update the scalar range on the actor's mapper
            actor.mapper.scalar_range = (0, vmax)
            p.render()

        # Enable point picking with the callback function
        p.enable_point_picking(callback=point_picker_callback, pickable_window=False,show_message='Right click to select a point')

        # Add a slider to control the scalar index dynamically
        slider_widget = p.add_slider_widget(
            slider_callback,         # Function to call when the slider moves
            rng=[0, self.eq_psd_multipoint[0].shape[1] - 1],  # Range of the slider
            value=freq_index,      # Default value
            title="Frequency",       # Title of the slider
            pointa=(.005, .1),        # Starting point of the slider in the plotter window
            pointb=(.45, .1),        # Ending point of the slider in the plotter window
            style='modern',          
        )

        slider_rep = slider_widget.GetRepresentation()

        # Hide the numerical display above the slider
        slider_rep.SetShowSliderLabel(False)

        # Add the Matplotlib chart to the plotter
        
        h_chart = pv.ChartMPL(f, size=(0.45, 0.5), loc=(0.005, 0.2))
        h_chart.background_color = (1.0, 1.0, 1.0, 0.4)
        p.add_chart(h_chart)

        # Show the plotter
        p.show(full_screen=True)





from ast import Raise
import numpy as np
from scipy import stats
from scipy import integrate
from scipy import special
import warnings

class Low(object):
    """Class for fatigue life estimation using frequency domain 
    method by Low[1].

    Notes
    -----
    Numerical implementation supports only integer values of  
    S-N curve parameter k (inverse slope). Due to approximation of
    large stress cycles through McLaurin series, sufficient engineering
    precision is up to k=6 [1].

    References
    ----------
    [1] Y.M.Low. A method for accurate estimation of the fatigue damage 
        induced by bimodal processes. Probabilistic Engineering Mechanics, 25(1):75-85, 2010
    [2] Aleš Zorman and Janko Slavič and Miha Boltežar. 
        Vibration fatigue by spectral methods—A review with open-source support, 
        Mechanical Systems and Signal Processing, 2023, 
        https://doi.org/10.1016/j.ymssp.2023.110149
    
    Example
    -------
    Import modules, define time- and frequency-domain data

    >>> import FLife
    >>> import pyExSi as es
    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> # time-domain data
    >>> N = 2 ** 16  # number of data points of time signal
    >>> fs = 2048  # sampling frequency [Hz]
    >>> t = np.arange(0, N) / fs  # time vector
    >>> # frequency-domain data
    >>> M = N // 2 + 1  # number of data points of frequency vector
    >>> freq = np.arange(0, M, 1) * fs / N  # frequency vector
    >>> PSD_lower = es.get_psd(freq, 20, 60, variance = 5)  # lower mode of random process
    >>> PSD_higher = es.get_psd(freq, 100, 120, variance = 2)  # higher mode of random process
    >>> PSD = PSD_lower + PSD_higher # bimodal one-sided flat-shaped PSD

    Get Gaussian stationary signal, instantiate SpectralData object and plot PSD

    >>> rg = np.random.default_rng(123) # random generator seed
    >>> x = es.random_gaussian(N, PSD, fs, rg) # Gaussian stationary signal
    >>> sd = FLife.SpectralData(input=x, dt=1/fs) # SpectralData instance
    >>> plt.plot(sd.psd[:,0], sd.psd[:,1]) 
    >>> plt.xlabel('Frequency [Hz]')
    >>> plt.ylabel('PSD')

    Define S-N curve parameters and get fatigue-life estimatate

    >>> C = 1.8e+22  # S-N curve intercept [MPa**k]
    >>> k = 7.3 # S-N curve inverse slope [/]
    >>> low = FLife.Low(sd, PSD_splitting=('userDefinedBands', [80,150]))
    >>> print(f'Fatigue life: {low.get_life(C,int(k)):.3e} s.')

    Plot segmentated PSD, used in Low's method

    >>> lower_band_index, upper_band_index= low.band_stop_indexes
    >>> plt.plot(sd.psd[:,0], sd.psd[:,1])
    >>> plt.vlines(sd.psd[:,0][lower_band_index], 0, np.max(sd.psd[:,1]), 'k', linestyles='dashed', alpha=.5)
    >>> plt.fill_between(sd.psd[:lower_band_index,0], sd.psd[:lower_band_index,1], 'o', label='lower band', alpha=.2, color='blue')
    >>> plt.vlines(sd.psd[:,0][upper_band_index], 0, np.max(sd.psd[:,1]), 'k', linestyles='dashed', alpha=.5)
    >>> plt.fill_between(sd.psd[lower_band_index:upper_band_index,0], sd.psd[lower_band_index:upper_band_index,1], 'o', label='upper band', alpha=.5, color ='orange')
    >>> plt.xlabel('Frequency [Hz]')
    >>> plt.ylabel('PSD')
    >>> plt.xlim(0,300)
    >>> plt.legend()
    """
    def __init__(self, spectral_data, PSD_splitting = ('equalAreaBands', 2)):
        """Get needed values from reference object.

        :param spectral_data: Instance of class SpectralData
        :param PSD_splitting: tuple
                PSD_splitting[0] is PSD spliting method, PSD_splitting[1] is method argument. 
                Splitting methods:

                - 'userDefinedBands', PSD_splitting[1] must be of type list or tupple, with N 
                  elements specifying upper band frequencies of N random processes.
                - 'equalAreaBands', PSD_splitting[1] must be of type int, specifying N random processes.

                Defaults to ('equalAreaBands', 2).
         """    
        self.spectral_data = spectral_data
        self.PSD_splitting = PSD_splitting
        self.band_stop_indexes = self.spectral_data._get_band_stop_frequency(self.PSD_splitting)

    def get_life(self, C, k):
        """Calculate fatigue life with parameters C, k, as defined in [2].

        :param C: [int,float]
            S-N curve intercept [MPa**k].
        :param k: [int,float]
            S-N curve inverse slope [/].
        :return:
            Estimated fatigue life in seconds.
        :rtype: float
        """
        #Data type check
        if not isinstance(k, int):
            raise Exception('Parameter `k` must be integer. Sufficient engineering precision is up to k=6.')
        
        if k > 6:
            warnings.warn(f'Sufficient engineering precision is up to k=6. Results should be evaluated carefully.')
        
        # -- spectral moments for each narrowband
        moments = self.spectral_data.get_spectral_moments(self.PSD_splitting, moments=[0])
        m0L, = moments[0] #spectral moments for lower band
        m0H, = moments[1] #spectral moments for upper band
        
        # -- positive slope zero crossing frequency
        v0L, v0H = self.spectral_data.get_nup(self.PSD_splitting)
        v0Small = v0H - v0L  #frequency of small cycless

        #band's zero crossing frequency ratio
        beta = v0H/v0L

        #Damage from small cycles
        #stress discrepancy
        eps = lambda r_lf, phi, beta: np.pi / (2*beta) * r_lf * np.abs(np.sin(phi))
        #peak PDF
        pdf_r = lambda r, var: stats.rayleigh.pdf(r, scale=np.sqrt(var))
        #phase angle
        pdf_phi = lambda phi, beta: stats.uniform.pdf(phi, loc=np.pi/4/beta, scale=np.pi/2-np.pi/4/beta)

        #damage
        int_func_small = lambda phi, r_lf: self._inner_integral_small(k, eps(r_lf, phi, beta), m0H) * pdf_phi(phi, beta) * pdf_r(r_lf, m0L)
        I_small = integrate.dblquad(int_func_small, 0, np.inf, lambda r_lf: np.pi/4/beta, lambda r_lf: np.pi/2)[0]
        d_small = v0Small / C * I_small

        #Damage from large cycles
        int_func_large = lambda r_hf, r_lf: self._inner_integral_large(k, r_lf, r_hf, beta) *  pdf_r(r_hf, m0H) * pdf_r(r_lf, m0L)
        I_large = 1/np.pi * integrate.dblquad(int_func_large, 0, np.inf, lambda r_lf: 0, lambda r_lf: np.inf)[0]
        d_large = v0L / C  * I_large 

        #Agregated damage
        d = d_small + d_large
        T = 1 / d
        return T

    def _Ik(self, eps, var, K):
        """Calculates coefficients for binomial series expansion for small cycles damage estimation[1].
        """
        Ik_array = np.zeros(K)
        Ik_array[0] = np.exp(-eps**2/(2*var))
        Ik_array[1] = eps * Ik_array[0] + np.sqrt(2*np.pi) * np.sqrt(var) * stats.norm.cdf(-eps/np.sqrt(var))
        for i in range(K-2):
            term1 = eps**(i+2) * Ik_array[0]
            term2 = (i+2) * var * Ik_array[i]
            Ik_array[i+2] = term1 + term2
        return Ik_array

    def _inner_integral_small(self, k, eps, var):
        """Returns analytical expression of innermost integral for small cycles damage estimation,
        based on binomial series expansion[1].
        """
        out = 0
        Ik_array = self._Ik(eps,var,int(k)+1)
        for i in range(len(Ik_array)):
            out += special.binom(k,i) * (-eps)**(k-i) * Ik_array[i]
        return out

    def _ro_j(self, r_lf, r_hf, beta, j):
        """Calculates coefficients for MacLaurin series expansion for large cycles damage estimation[1].
        """
        c = beta * r_hf / (r_lf + beta**2 * r_hf)
        out = r_lf * c**j + r_hf * (beta*c -1)**j
        out /= special.factorial(j)
        return out

    def _inner_integral_large(self, k, r_lf, r_hf, beta):
        """Returns analytical approximation of innermost integral for large cycles damage estimation,
        based on MacLaurin series expansion. The approximation gives sufficient engineering
        precision in the damage estimate for up to k = 6 [1].
        """
        r_sum = r_lf + r_hf
        ro_2 = self._ro_j(r_lf, r_hf, beta, 2)
        ro_4 = self._ro_j(r_lf, r_hf, beta, 4)
        ro_6 = self._ro_j(r_lf, r_hf, beta, 6)
        
        out = r_sum**k * (np.pi - 1/3/r_sum * k * ro_2 * np.pi**3 \
            + k/5/r_sum * (ro_4 + (k-1) * ro_2**2 /2/r_sum) * np.pi**5 \
            - k/7/r_sum * (ro_6 + (k-1) * ro_2 * ro_4 /r_sum + (k-1) * (k-2) * ro_2**3 /(6*r_sum**2)) * np.pi**7)
        return out


import numpy as np
from .narrowband import Narrowband

class BandsMethod(Narrowband):
    """Class for fatigue life estimation using frequency domain 
    method by Braccesi et al. [1].
    
    References
    ----------
    [1] Claudio Braccesi, Filippo Cianetti and Lorenzo Tomassini. Random fatigue. A new 
        frequency domain criterion for the damage evaluation of mechanical components.
        International Journal of Fatigue, 70:417-427, 2015
    [2] Aleš Zorman and Janko Slavič and Miha Boltežar. 
        Vibration fatigue by spectral methods—A review with open-source support, 
        Mechanical Systems and Signal Processing, 2023, 
        https://doi.org/10.1016/j.ymssp.2023.110149
    
    Example
    -------
    Import modules, define time- and frequency-domain data

    >>> import FLife
    >>> import pyExSi as es
    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> # time-domain data
    >>> N = 2 ** 16  # number of data points of time signal
    >>> fs = 2048  # sampling frequency [Hz]
    >>> t = np.arange(0, N) / fs  # time vector
    >>> # frequency-domain data
    >>> M = N // 2 + 1  # number of data points of frequency vector
    >>> freq = np.arange(0, M, 1) * fs / N  # frequency vector
    >>> PSD_lower = es.get_psd(freq, 20, 60, variance = 5)  # lower mode of random process
    >>> PSD_higher = es.get_psd(freq, 100, 120, variance = 2)  # higher mode of random process
    >>> PSD = PSD_lower + PSD_higher # bimodal one-sided flat-shaped PSD

    Get Gaussian stationary signal, instantiate SpectralData object and plot PSD

    >>> rg = np.random.default_rng(123) # random generator seed
    >>> x = es.random_gaussian(N, PSD, fs, rg) # Gaussian stationary signal
    >>> sd = FLife.SpectralData(input=x, dt=1/fs) # SpectralData instance
    >>> plt.plot(sd.psd[:,0], sd.psd[:,1]) 
    >>> plt.xlabel('Frequency [Hz]')
    >>> plt.ylabel('PSD')

    Define S-N curve parameters and get fatigue-life estimatate

    >>> C = 1.8e+22  # S-N curve intercept [MPa**k]
    >>> k = 7.3 # S-N curve inverse slope [/]
    >>> bm = FLife.BandsMethod(sd, PSD_splitting=('userDefinedBands', [80,150]))
    >>> print(f'Fatigue life: {bm.get_life(C,k):.3e} s.')

    Plot segmentated PSD, used in Bands method

    >>> lower_band_index, upper_band_index= bm.band_stop_indexes
    >>> plt.plot(sd.psd[:,0], sd.psd[:,1])
    >>> plt.vlines(sd.psd[:,0][lower_band_index], 0, np.max(sd.psd[:,1]), 'k', linestyles='dashed', alpha=.5)
    >>> plt.fill_between(sd.psd[:lower_band_index,0], sd.psd[:lower_band_index,1], 'o', label='lower band', alpha=.2, color='blue')
    >>> plt.vlines(sd.psd[:,0][upper_band_index], 0, np.max(sd.psd[:,1]), 'k', linestyles='dashed', alpha=.5)
    >>> plt.fill_between(sd.psd[lower_band_index:upper_band_index,0], sd.psd[lower_band_index:upper_band_index,1], 'o', label='upper band', alpha=.5, color ='orange')
    >>> plt.xlabel('Frequency [Hz]')
    >>> plt.ylabel('PSD')
    >>> plt.xlim(0,200)
    >>> plt.legend()
    """
    def __init__(self, spectral_data, PSD_splitting = ('equalAreaBands', 1)):
        """Get needed values from reference object.

        :param spectral_data:  Instance of class SpectralData       
        :param PSD_splitting: tuple
                PSD_splitting[0] is PSD spliting method, PSD_splitting[1] is method argument. 
                Splitting methods:

                - 'userDefinedBands', PSD_splitting[1] must be of type list or tupple, with N 
                  elements specifying upper band frequencies of N random processes.
                - 'equalAreaBands', PSD_splitting[1] must be of type int, specifying N random processes.

                Defaults to ('equalAreaBands', 1).
        """
        Narrowband.__init__(self, spectral_data)
        self.PSD_splitting = PSD_splitting
        self.band_stop_indexes = self.spectral_data._get_band_stop_frequency(self.PSD_splitting)
        
    def get_life(self, C, k):
        """Calculate fatigue life with parameters C, k, as defined in [1, 2].

        :param C: [int,float]
            S-N curve intercept [MPa**k].
        :param k: [int,float]
            S-N curve inverse slope [/].
        :return:
            Estimated fatigue life in seconds.
        :rtype: float
        """
        #Reference frequency (is arbitrary and is taken as expected frequency of positive 
        #slope zero crossing of random process (described by SpectralData instance))
        f_ref, = self.spectral_data.get_nup(self.spectral_data.PSD_splitting) 

        if len(self.band_stop_indexes) == 1: # modified PSD formulation
            m0_ref = self._m0_ref_modifiedPSD(k,f_ref)
        else: # user defined bands
            m0_ref = self._m0_ref_userDefinedBands(k,f_ref)

        d = self.damage_intesity_NB(m0=m0_ref, nu=f_ref, C=C, k=k) 
        T = 1/d     
        return T


    def _m0_ref_userDefinedBands(self, k, f_ref):
        # expected frequency of positive slope zero crossing
        bands_central_freq = self.spectral_data.get_nup(self.PSD_splitting)
        # bands variance
        m0_array = self.spectral_data.get_spectral_moments(self.PSD_splitting, moments=[0])[:,0]

        f = self.spectral_data.psd[:, 0]
        m0_ref_list = []
        for m0, band_freq in zip(m0_array, bands_central_freq):
            band_freq = f[np.abs(f - band_freq).argmin()]
            mo_ref = (band_freq / f_ref)**(2/k) * m0
            m0_ref_list.append(mo_ref)
        m0_ref_list = np.array(m0_ref_list)

        m0_ref = np.sum(m0_ref_list)
        return m0_ref

    def _m0_ref_modifiedPSD(self, k, f_ref):
        f = self.spectral_data.psd[:, 0]
        PSD = self.spectral_data.psd[:, 1] 
        modified_PSD = []
        df = f[1] - f[0] 

        for f_i, PSD_i in zip(f,PSD):
            PSD_i_modified = (f_i / f_ref)**(2/k) * PSD_i
            modified_PSD.append(PSD_i_modified)
        modified_PSD = np.array(modified_PSD)

        if np.__version__>='2.0.0':
            trapezoid = np.trapezoid
        else:
            trapezoid = np.trapz
            
        m0_ref = trapezoid(modified_PSD, dx=df)
        return m0_ref

    def get_PDF(self, s):
        raise Exception(f'Function <get_PDF> is not available for class {self.__class__.__name__:s}.')


import numpy as np
from . import cplane
from scipy.optimize import minimize

def _max_normal(self, s, search_method):
    """
    Internal function for calculating equivalent stress at one node.
    """
    
    freq = self.multiaxial_psd[1]
    df = freq[1] - freq[0]
    
    l1, m1, n1, l2, m2, n2, l3, m3, n3 = cplane.max_variance(multiaxial_psd=s, df=df, method='maxnormal',search_method=search_method)
    
    a = np.asarray([l1**2, m1**2, n1**2,
                2*l1*m1, 2*l1*n1, 2*m1*n1])

    s_eq = np.einsum('i,zij,j', a, s, a)
    return s_eq


def _max_shear(self, s, search_method):
    """
    Internal function for calculating equivalent stress at one node.
    """
    
    freq = self.multiaxial_psd[1]
    df = freq[1] - freq[0]

    l1, m1, n1, l2, m2, n2, l3, m3, n3 = cplane.max_variance(multiaxial_psd=s, df=df, method='maxshear', search_method=search_method)

    a = np.asarray([l1**2-l3**2, m1**2-m3**2, n1**2-n3**2,
            2*(l1*m1-l3*m3), 2*(l1*n1-l3*n3), 2*(m1*n1-m3*n3)])

    s_eq = np.einsum('i,zij,j', a, s, a)
    return s_eq


def _max_normal_and_shear(self, s, s_af, tau_af, search_method='local'):
    """
    Internal function for calculating equivalent stress at one node.
    """
    
    freq = self.multiaxial_psd[1]
    df = freq[1] - freq[0]

    K = 2 * tau_af / s_af - 1
    l1, m1, n1, l2, m2, n2, l3, m3, n3 = cplane.max_variance(multiaxial_psd=s, df=df, method='maxshear',K=K, search_method=search_method)
    
    a = np.asarray([l1**2-l3**2+K*(l1+l3)**2,
                    m1**2-m3**2+K*(m1+m3)**2,
                    n1**2-n3**2+K*(n1+n3)**2,
                    2*(l1*m1-l3*m3+K*(l1+l3)*(m1+m3)),
                    2*(l1*n1-l3*n3+K*(l1+l3)*(n1+n3)),
                    2*(m1*n1-m3*n3+K*(m1+m3)*(n1+n3))]) / (1 + K) 

    s_eq = np.einsum('i,zij,j', a, s, a)
    return s_eq


def _EVMS(self, s):
    """
    Internal function for calculating equivalent stress at one node.
    """
    
    freq = self.multiaxial_psd[1]
    df = freq[1] - freq[0]

    Q = np.array([
        [1,-1/2,-1/2,0,0,0],
        [-1/2,1,-1/2,0,0,0],
        [-1/2,-1/2,1,0,0,0],
        [0,0,0,3,0,0],
        [0,0,0,0,3,0],
        [0,0,0,0,0,3]])
    
    Q_2D = np.array([
        [1,-1/2,0],
        [-1/2,1,0],
        [0,0,3],])

    if s.shape[1] == 6 and s.shape[2] == 6:
        s_eq = np.trace(np.einsum('ij,kjl->kil', Q, s),axis1=1,axis2=2)
    elif s.shape[1] == 3 and s.shape[2] == 3:
        s_eq = np.trace(np.einsum('ij,kjl->kil', Q_2D, s),axis1=1,axis2=2)
    else:
        raise Exception('Input Error. PSD matrix should be the size of (f,6,6) for 3D stress state or (f,3,3) for 2D stress state')
    return s_eq


def _cs(self, s, s_af, tau_af):
    """
    Internal function for calculating equivalent stress at one node.
    """
    
    freq = self.multiaxial_psd[1]
    df = freq[1] - freq[0]

    # .. Find critical plane.
    C = cplane.csrandom(multiaxial_psd=s, df=df, s_af=s_af, tau_af=tau_af)
    
    # .. Transform stress.
    s = np.einsum('ij,zjk,lk->zil', C, s, C)
    
    # .. Calculate equivalent PSD.
    sww = s[:, 2, 2]
    suw = s[:, -1, -1,]

    s_eq = sww.real + suw.real * (s_af/tau_af)
    return s_eq


def _multiaxial_rainflow(self, s):
    """
    Internal function for calculating equivalent stress at one node.
    """
    if s.shape[1] == 6 and s.shape[2] == 6:
        raise ValueError('Multiaxial rainflow only works for biaxial stresses: PSD matrix (f,3,3)')
    
    freq = self.multiaxial_psd[1]
    df = freq[1] - freq[0]

    

    def constraint(c):
        return np.sum(c**2) - 1

    initial_guess = np.array([0.34,0.33,0.33])

    # Define the constraint for the optimizaton
    cons = {'type': 'eq', 'fun': constraint}

    # Perform the optimization
    s_eq = np.empty((len(s)))
    for i in range(len(s)):
        def Psi_m(c):
            Q = np.outer(c, c)
            return -np.trace(Q@s[i])
        
        result = minimize(fun=Psi_m, x0=initial_guess, constraints=[cons], method='trust-constr')


        if not result.success:
            print("Optimization failed:", result.message)

            s_eq[i] -Psi_m(initial_guess)  
        # Extract the optimized c values
        else:    
            c_opt = result.x

            s_eq[i] = -Psi_m(c_opt)
    return s_eq


def _thermoelastic(self,s):
    """
    Internal function for calculating equivalent stress at one node.
    """
    
    freq = self.multiaxial_psd[1]
    df = freq[1] - freq[0]

    Q = np.array([
        [1,1,1,0,0,0],
        [1,1,1,0,0,0],
        [1,1,1,0,0,0],
        [0,0,0,0,0,0],
        [0,0,0,0,0,0],
        [0,0,0,0,0,0]])

    Q_2D = np.array([
        [1,1,0],
        [1,1,0],
        [0,0,0]])
    
    if s.shape[1] == 6 and s.shape[2] == 6:
        s_eq = np.trace(np.einsum('ij,kjl->kil', Q, s),axis1=1,axis2=2)
    elif s.shape[1] == 3 and s.shape[2] == 3:
        s_eq = np.trace(np.einsum('ij,kjl->kil', Q_2D, s),axis1=1,axis2=2)
    else:
        raise Exception('Input Error. PSD matrix should be the size of (f, 6, 6) for 3D stress state or (f,3,3) for 2D stress state')
                
    return s_eq


def _liwi(self,s):
    """
    Internal function for calculating equivalent stress at one node.
    """
    s_eq = np.empty((len(s)),dtype=complex)

    for i in range(len(s)):
        sigma_x, sigma_y, tau_xy = s[i,0], s[i,1], s[i,2]

        # Calculate absolute values and angles
        sigma_x_abs = np.abs(sigma_x)
        sigma_y_abs = np.abs(sigma_y)
        tau_xy_abs = np.abs(tau_xy)

        sigma_x_angle = np.angle(sigma_x)
        sigma_y_angle = np.angle(sigma_y)
        tau_xy_angle = np.angle(tau_xy)

        # Define initial parameters
        A1 = sigma_x_abs
        phi_1 = sigma_x_angle
        r = 0

        A2 = np.array([sigma_y_abs, np.sqrt(3) * tau_xy_abs])
        phi_2 = np.array([sigma_y_angle, tau_xy_angle])
        theta = np.array([np.arccos(-1/2), np.pi / 2])

        while r < 2:
            # Calculate intermediate variables in the main loop
            phi_a = np.angle(A1 + A2[r] * np.e**(1j * (phi_2[r] - phi_1 - theta[r])))
            phi_c_overline = theta[r] - 0.5 * np.angle(A2[r]**2 + 2 * A1 * A2[r] * np.cos(phi_1 - phi_2[r]) * np.e**(1j * theta[r]) + A1**2 * np.e**(2j * theta[r]))

            if A2[r] >= A1:
                phi_c = ((phi_c_overline - theta[r] / 2) % (np.pi/2)) + theta[r] / 2
            elif A2[r] < A1:
                phi_c = ((phi_c_overline - theta[r] / 2) % (np.pi/2)) + theta[r] / 2 - np.pi / 2
            else:
                phi_c = phi_c_overline

            A1 = (0.5 * (A1**2 + A2[r]**2 + 2 * A1 * A2[r] * np.cos(phi_1 - phi_2[r]) * np.cos(theta[r])) + 0.5 * np.abs(A1**2 * np.e**(2j * phi_1) + A2[r]**2 * np.e**(2j * phi_2[r]) + 2 * A1 * A2[r] * np.e**(2j * (phi_1 + phi_2[r]))))**0.5
                    
            phi_1 = phi_a + phi_c + phi_1
            r += 1  # Increment r

        # Calculate phi_B0
        phi_B0 = 0.5 * np.angle(sigma_x_abs**2 * np.e**(2j * sigma_x_angle) + sigma_y_abs**2 * np.e**(2j * sigma_y_angle) - sigma_x_abs * sigma_y_abs * np.e**(1j * (sigma_x_angle + sigma_y_angle)) + 3 * tau_xy_abs**2 * np.e**(2j * tau_xy_angle))

        if np.abs(np.angle((np.e**(1j * phi_1))/(np.e**(1j * phi_B0)))) < np.abs(np.angle((np.e**(1j * phi_1))/(np.e**(1j * (phi_B0+np.pi))))):
            phi_vM = phi_B0
        else:
            phi_vM = phi_B0 + np.pi


        # Calculate delta_UM
        sigma_vM_abs = (0.5 * np.abs(sigma_x_abs**2 + sigma_y_abs**2 - sigma_x_abs * sigma_y_abs * np.cos(sigma_x_angle - sigma_y_angle) + 3 * tau_xy_abs**2) + 0.5 * np.abs(sigma_x_abs**2 * np.e**(2j * sigma_x_angle) + sigma_y_abs**2 * np.e**(2j * sigma_y_angle) - sigma_x_abs * sigma_y_abs * np.e**(1j*(sigma_x_angle+sigma_y_angle)) + 3 * tau_xy_abs**2 * np.e**(2j * tau_xy_angle)))**0.5

        sigma_vM = sigma_vM_abs * np.e**(1j * phi_vM)

        s_eq[i] = sigma_vM
    
    return s_eq


def _coin_liwi(self, s, k_a, k_phi):
    """
    Internal function for calculating equivalent stress at one node.
    """
    s_eq = np.empty((len(s)),dtype=complex)

    for i in range(len(s)):

        sigma_x_f = s[i,0]
        sigma_y_f = s[i,1]  
        sigma_z_f = s[i,2]  
        tau_xy_f = s[i,3]
        tau_xz_f = s[i,4]   
        tau_yz_f = s[i,5]   

        S = np.array([[sigma_x_f, tau_xy_f, tau_xz_f],
                    [tau_xy_f, sigma_y_f, tau_yz_f],
                    [tau_xz_f, tau_yz_f, sigma_z_f]])

        I_1_underscore = np.trace(S)

        I_1_squared = np.trace(S) * np.conj(np.trace(S))

        I_1_squared_underscore = np.trace(S) * np.trace(S)

        I_2 = 0.5 * (I_1_squared - np.trace(S@np.conj(S)))

        I_2_underscore = 0.5 * (I_1_squared_underscore - np.trace(S @ S))



        sigma_eq_ka = I_1_squared - k_a**2 * I_2

        sigma_eq_ka_underscore = I_1_squared_underscore - k_a**2 * I_2_underscore

        phi_I1 = np.angle(I_1_underscore)

        phi_eq_ka = np.angle(np.sqrt(sigma_eq_ka_underscore))


        # amplitude
        sigma_coin = np.sqrt(k_phi * sigma_eq_ka + (1 - k_phi) * abs(sigma_eq_ka_underscore))

        #phase angle

        if np.angle(np.exp(1j * phi_I1) / np.exp(1j * phi_eq_ka)) < np.angle(np.exp(1j * phi_I1) / np.exp(1j * phi_eq_ka + np.pi)):
            phi_coin = phi_eq_ka

        else:
            phi_coin = phi_eq_ka + np.pi 

        sigma_coin_f = sigma_coin * np.exp(1j * phi_coin)

        s_eq[i] = sigma_coin_f
    
    return s_eq

def _EVMS_out_of_phase(self, s):
    """
    Internal function for calculating equivalent stress at one node.
    """
    s_eq = np.empty((len(s)))

    for i in range(len(s)):
        sigma_xx_squared = s[i,0,0]
        sigma_yy_squared = s[i,1,1]
        tau_xy_squared = s[i,2,2]
        sigma_xx_sigma_yy = np.sqrt(sigma_xx_squared * sigma_yy_squared)

        phi_yy_phi_xx = np.angle(s[i,0,1])
        phi_xx_phi_yy = np.angle(s[i,1,0])
        phi_xy_phi_xx = np.angle(s[i,0,2])

        s_eq[i] = 0.5  * (sigma_xx_squared + sigma_yy_squared - sigma_xx_sigma_yy * np.cos(phi_xx_phi_yy) + 3 * tau_xy_squared) + 0.5 * np.abs(sigma_xx_squared + sigma_yy_squared * np.exp(2j * phi_yy_phi_xx) - sigma_xx_sigma_yy * np.exp(1j * phi_yy_phi_xx) + 3*tau_xy_squared * np.exp(2j * phi_xy_phi_xx))

    return s_eq

def _Nieslony(self, s, s_af, tau_af,coefficient_load_type):
    """
    Internal function for calculating equivalent stress at one node.
    """

    G_p = 1/9*_thermoelastic(self,s)
    G_EVMS = _EVMS(self,s)

    s_eq = np.empty((len(s)))

    for i in range(len(s)): 
        a_sigma = np.sqrt(2*np.diagonal(s[i]))
        #find the index of maximum value of a_sigma
        k = np.argmax(a_sigma)
        phi_k = np.angle(s[i,:,k])
        s_ = a_sigma * np.exp(1j * phi_k)
        sigma_EMSc = 1/np.sqrt(2) * np.sqrt((s_[0]-s_[1])**2 + (s_[1]-s_[2])**2 + (s_[2]-s_[0])**2 + 6*(s_[3]**2+s_[4]**2+s_[5]**2))
        sigma_pc = 1/3 * (s_[0]+s_[1]+s_[2])
        rp_EMS = np.abs(np.cos(np.angle(sigma_EMSc)-np.angle(sigma_pc)))

        if coefficient_load_type == 'tension':
            A = (-np.sqrt(3) / tau_af) * ((s_af * rp_EMS) - np.sqrt((s_af ** 2) * (rp_EMS ** 2) - (s_af ** 2) + 3 * (tau_af ** 2)))
            B = (np.sqrt(3) / 3) * (s_af / tau_af)

        elif coefficient_load_type == 'torsion':
            A = (-np.sqrt(3) / s_af) * ((s_af * rp_EMS) - np.sqrt((s_af ** 2) * (rp_EMS ** 2) - (s_af ** 2) + 3 * (tau_af ** 2)))
            B = np.sqrt(3) / 3
        else:
            raise ValueError('Invalid coefficient_load_type. Please choose either "tension" or "torsion"')
        
        s_eq[i] = A**2 * G_p[i] + B**2 * G_EVMS[i] + 2 * A * B * rp_EMS * np.sqrt(G_p[i] * G_EVMS[i])
    
    return s_eq

def _Lemaitre(self, s, poisson_ratio):
    """
    Internal function for calculating equivalent stress at one node.
    """
    
    freq = self.multiaxial_psd[1]
    df = freq[1] - freq[0]

    Q = np.array([
        [1,-poisson_ratio,-poisson_ratio,0,0,0],
        [-poisson_ratio,1,-poisson_ratio,0,0,0],
        [-poisson_ratio,-poisson_ratio,1,0,0,0],
        [0,0,0,2*(1+poisson_ratio),0,0],
        [0,0,0,0,2*(1+poisson_ratio),0],
        [0,0,0,0,0,2*(1+poisson_ratio)]])
    
    Q_2D = np.array([
        [1,-poisson_ratio,0],
        [-poisson_ratio,1,0],
        [0,0,2*(1+poisson_ratio)],])

    if s.shape[1] == 6 and s.shape[2] == 6:
        s_eq = np.trace(np.einsum('ij,kjl->kil', Q, s),axis1=1,axis2=2)
    elif s.shape[1] == 3 and s.shape[2] == 3:
        s_eq = np.trace(np.einsum('ij,kjl->kil', Q_2D, s),axis1=1,axis2=2)
    else:
        raise Exception('Input Error. PSD matrix should be the size of (f,6,6) for 3D stress state or (f,3,3) for 2D stress state')
    return s_eq


import numpy as np
from scipy import stats
from scipy import integrate
from .narrowband import Narrowband
from ..tools import pdf_rayleigh_sum

class ModifiedFuCebon(Narrowband):
    """Class for fatigue life estimation using frequency domain 
    method by Benasciutti and Tovo[1].
    
    References
    ----------
    [1] Denis Benasciutti and Roberto Tovo. Comparison of spectral methods for fatigue damage
        assessment in bimodal random processes. 9th International Conference
        on Structural Safety & Reliability (ICOSSAR), 230:3207-3214, 2005
    [2] Aleš Zorman and Janko Slavič and Miha Boltežar. 
        Vibration fatigue by spectral methods—A review with open-source support, 
        Mechanical Systems and Signal Processing, 2023, 
        https://doi.org/10.1016/j.ymssp.2023.110149
    
    Example
    -------
    Import modules, define time- and frequency-domain data

    >>> import FLife
    >>> import pyExSi as es
    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> # time-domain data
    >>> N = 2 ** 16  # number of data points of time signal
    >>> fs = 2048  # sampling frequency [Hz]
    >>> t = np.arange(0, N) / fs  # time vector
    >>> # frequency-domain data
    >>> M = N // 2 + 1  # number of data points of frequency vector
    >>> freq = np.arange(0, M, 1) * fs / N  # frequency vector
    >>> PSD_lower = es.get_psd(freq, 20, 60, variance = 5)  # lower mode of random process
    >>> PSD_higher = es.get_psd(freq, 100, 120, variance = 2)  # higher mode of random process
    >>> PSD = PSD_lower + PSD_higher # bimodal one-sided flat-shaped PSD

    Get Gaussian stationary signal, instantiate SpectralData object and plot PSD

    >>> rg = np.random.default_rng(123) # random generator seed
    >>> x = es.random_gaussian(N, PSD, fs, rg) # Gaussian stationary signal
    >>> sd = FLife.SpectralData(input=x, dt=1/fs) # SpectralData instance
    >>> plt.plot(sd.psd[:,0], sd.psd[:,1]) 
    >>> plt.xlabel('Frequency [Hz]')
    >>> plt.ylabel('PSD')

    Define S-N curve parameters and get fatigue-life estimatate

    >>> C = 1.8e+22  # S-N curve intercept [MPa**k]
    >>> k = 7.3 # S-N curve inverse slope [/]
    >>> mfc = FLife.ModifiedFuCebon(sd, PSD_splitting=('userDefinedBands', [80,150]))
    >>> print(f'Fatigue life: {mfc.get_life(C,k):.3e} s.')

    Plot segmentated PSD, used in modified Fu-Cebon method

    >>> lower_band_index, upper_band_index= mfc.band_stop_indexes
    >>> plt.plot(sd.psd[:,0], sd.psd[:,1])
    >>> plt.vlines(sd.psd[:,0][lower_band_index], 0, np.max(sd.psd[:,1]), 'k', linestyles='dashed', alpha=.5)
    >>> plt.fill_between(sd.psd[:lower_band_index,0], sd.psd[:lower_band_index,1], 'o', label='lower band', alpha=.2, color='blue')
    >>> plt.vlines(sd.psd[:,0][upper_band_index], 0, np.max(sd.psd[:,1]), 'k', linestyles='dashed', alpha=.5)
    >>> plt.fill_between(sd.psd[lower_band_index:upper_band_index,0], sd.psd[lower_band_index:upper_band_index,1], 'o', label='upper band', alpha=.5, color ='orange')
    >>> plt.xlabel('Frequency [Hz]')
    >>> plt.ylabel('PSD')
    >>> plt.xlim(0,300)
    >>> plt.legend()
    """
    def __init__(self, spectral_data, PSD_splitting = ('equalAreaBands', 2)):
        """Get needed values from reference object.

        :param spectral_data: Instance of class SpectralData
        :param PSD_splitting: tuple
                PSD_splitting[0] is PSD spliting method, PSD_splitting[1] is method argument. 
                Splitting methods:

                - 'userDefinedBands', PSD_splitting[1] must be of type list or tupple, with N 
                  elements specifying upper band frequencies of N random processes.
                - 'equalAreaBands', PSD_splitting[1] must be of type int, specifying N random processes.

                Defaults to ('equalAreaBands', 2).
         """     
        Narrowband.__init__(self, spectral_data)
        self.PSD_splitting = PSD_splitting
        self.band_stop_indexes = self.spectral_data._get_band_stop_frequency(self.PSD_splitting)

    def get_life(self, C, k):
        """Calculate fatigue life with parameters C, k, as defined in [2].

        :param C: [int,float]
            S-N curve intercept [MPa**k].
        :param k: [int,float]
            S-N curve inverse slope [/].
        :return:
            Estimated fatigue life in seconds.
        :rtype: float
        """
        # -- spectral moments for each narrowband
        moments = self.spectral_data.get_spectral_moments(self.PSD_splitting, moments=[0,2])
        m0L, m2L = moments[0] #spectral moments for lower band
        m0H, m2H = moments[1] #spectral moments for upper band

        # -- Vanmarcke bandwidth parameter
        _, epsV_H = self.spectral_data.get_vanmarcke_parameter(self.PSD_splitting)

        # -- positive slope zero crossing frequency
        v0L, v0H = self.spectral_data.get_nup(self.PSD_splitting)

        # -- normalized variances
        m0 = np.sum(moments[:, 0])
        m0L_norm = m0L/m0
        m0H_norm = m0H/m0

        v0Large = m0L_norm * v0L* np.sqrt(1 + m0H_norm/m0L_norm * (v0H/v0L*epsV_H)**2) #low + high frequency, large amplitudes
        v0Small = v0H - v0Large  #freqeuncy of small cycless

        #dNB small
        #small cycles consist of high frequency component
        dNB_small = self.damage_intesity_NB(m0H, v0Small, C, k) 

        #dNB large
        #large cycles consist of low and high frequency component
        pdf_large = pdf_rayleigh_sum(m0L,m0H)
        S_large = integrate.quad(lambda x:  x**k * pdf_large(x), 0, np.inf)[0]
        dNB_large = v0Large * S_large / C
        d = dNB_small + dNB_large
        T = 1 / d
        return T

    def get_PDF(self, s):
        raise Exception(f'Function <get_PDF> is not available for class {self.__class__.__name__:s}.')


import numpy as np
from scipy.integrate import quad
from scipy.special import gamma
import warnings

class JunPark(object):
    """Class for fatigue life estimation using frequency domain 
    method by Jun and Park[1].
    
    References
    ----------
    [1] Seock-Hee Jun and Jun-Bum Park. Development of a novel fatigue damage model for 
        Gaussian wide band stress responses using numerical approximation methods. International 
        Journal of Naval Architecture and Ocean Engineering, 12: 755-767, 2020
    [2] Aleš Zorman and Janko Slavič and Miha Boltežar. 
        Vibration fatigue by spectral methods—A review with open-source support, 
        Mechanical Systems and Signal Processing, 2023, 
        https://doi.org/10.1016/j.ymssp.2023.110149
        
    Example
    -------
    Import modules, define time- and frequency-domain data

    >>> import FLife
    >>> import pyExSi as es
    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> # time-domain data
    >>> N = 2 ** 16  # number of data points of time signal
    >>> fs = 2048  # sampling frequency [Hz]
    >>> t = np.arange(0, N) / fs  # time vector
    >>> # frequency-domain data
    >>> M = N // 2 + 1  # number of data points of frequency vector
    >>> freq = np.arange(0, M, 1) * fs / N  # frequency vector
    >>> PSD_lower = es.get_psd(freq, 20, 60, variance = 5)  # lower mode of random process
    >>> PSD_higher = es.get_psd(freq, 100, 120, variance = 2)  # higher mode of random process
    >>> PSD = PSD_lower + PSD_higher # bimodal one-sided flat-shaped PSD

    Get Gaussian stationary signal, instantiate SpectralData object and plot PSD

    >>> rg = np.random.default_rng(123) # random generator seed
    >>> x = es.random_gaussian(N, PSD, fs, rg) # Gaussian stationary signal
    >>> sd = FLife.SpectralData(input=x, dt=1/fs) # SpectralData instance
    >>> plt.plot(sd.psd[:,0], sd.psd[:,1]) 
    >>> plt.xlabel('Frequency [Hz]')
    >>> plt.ylabel('PSD')

    Define S-N curve parameters and get fatigue-life estimatate

    >>> C = 1.8e+22  # S-N curve intercept [MPa**k]
    >>> k = 7.3 # S-N curve inverse slope [/]
    >>> jp = FLife.Jun(sd)
    >>> print(f'Fatigue life: {jp.get_life(C,k):.3e} s.')

    Define stress vector and depict stress peak PDF

    >>> s = np.arange(0,np.max(x),.01) 
    >>> plt.plot(s,jp.get_PDF(s))
    >>> plt.xlabel('Stress [MPa]')
    >>> plt.ylabel('PDF')
    """
    def __init__(self, spectral_data):
        """Get needed values from reference object.

        :param spectral_data:  Instance of class SpectralData
        """     
        self.spectral_data = spectral_data
        self._set_distribution_parameters() #calculate distribution parameters

    def get_PDF(self, s):
        """Returns cycle PDF(Probability Density Function) as a function of stress s.

        :param s:  numpy.ndarray
            Stress vector.
        :return: function pdf(s)
        """
        m0 = self.spectral_data.moments[0]
        Qc = self._get_Qc() # PDF correction factor Qc

        def jun_pdf(s):
            #PDF of stress amplitude normalized by standard deviation of process
            #exponential
            exponential_pdf = lambda s: 1/self.parameters['sigma_E'] * np.exp(-s/self.parameters['sigma_E'])
            #Rayleigh
            rayleigh1_pdf = lambda s: s/self.parameters['sigma_R']**2 * np.exp(-s**2/(2*self.parameters['sigma_R']**2))
            #Rayleigh with unit variance
            rayleigh2_pdf = lambda s: s * np.exp(-s**2/2)
            #half-Gaussian
            gauss_pdf = lambda s: 2/(np.sqrt(2*np.pi)*self.parameters['sigma_H'])* np.exp(-s**2/(2*self.parameters['sigma_H']**2))

            pdf = self.parameters['D_1']*exponential_pdf(s) + self.parameters['D_2']*rayleigh1_pdf(s) \
                + self.parameters['D_3']*rayleigh2_pdf(s) + self.parameters['D_4']*gauss_pdf(s)
            return  Qc * pdf
        
        return 1/np.sqrt(m0) * jun_pdf(s/np.sqrt(m0))

    def get_life(self, C, k, integrate_pdf=False):
        """Calculate fatigue life with parameters C, k, as defined in [2].

        :param C: [int,float]
            S-N curve intercept [MPa**k].
        :param k: [int,float]
            S-N curve inverse slope [/].
        :return:
            Estimated fatigue life in seconds.
        :rtype: float
        """ 
        m0 = self.spectral_data.moments[0]
        m_p = self.spectral_data.m_p

        if integrate_pdf:
            d = m_p / C * quad(lambda s: s**k*self.get_PDF(s), a=0, b=np.inf)[0]
        else:
            Qc = self._get_Qc()
            d = Qc * m_p / C * (np.sqrt(2*m0))**k * (self.parameters['D_1']/(np.sqrt(2)**k)* self.parameters['sigma_E']**k * gamma(1+k) \
                + self.parameters['D_2']*self.parameters['sigma_R']**k * gamma(1 + k/2) + self.parameters['D_3']*gamma(1+k/2) \
                + self.parameters['D_4']/(np.sqrt(np.pi)) * self.parameters['sigma_H']**k * gamma((1+k)/2) )

        T = 1.0/d
        return T

    def _set_distribution_parameters(self):
        '''Define PDF parameters; 
        '''
        #Parameters for n-th moment of rainflow range distrubution Mrr(n)
        alpha_1 = self.spectral_data.alpha1
        alpha_2 = self.spectral_data.alpha2
        rho = alpha_1**1.1 * alpha_2**0.9

        # Define special bandwidth parameter mu_k [1]
        def get_mu_k(k):
            nominator = self.spectral_data.get_spectral_moments(self.spectral_data.PSD_splitting, moments=[k+0.01])[0][0]
            denominator = np.sqrt(self.spectral_data.get_spectral_moments(self.spectral_data.PSD_splitting, moments=[0.01])[0][0]\
                 * self.spectral_data.get_spectral_moments(self.spectral_data.PSD_splitting, moments=[2*k+0.01])[0][0])
            return nominator/denominator

        mu_1 = get_mu_k(1)
        mu_0_52 = get_mu_k(0.52)

        #Mrr(n)
        MRR_1 = rho * mu_1**-0.96
        MRR_2 = rho * mu_1**-0.02
        MRR_3 = rho * mu_0_52

        #distribution parameters
        sigma_R = alpha_2
        D_1 = 2*(alpha_1*alpha_2 - alpha_2**2)/(1 + alpha_2**2)
        D_2 = (MRR_2 - MRR_3)/(sigma_R**2*(1 - sigma_R))
        D_3 = (-sigma_R*MRR_2 + MRR_3)/(1 - sigma_R)
        D_4 = 1 - D_1 - D_2 - D_3

        A_1 = gamma(2)/(np.sqrt(2) * gamma(1.5))
        B_1 = 1/np.sqrt(np.pi) * gamma(1)/gamma(1.5)

        sigma_H = 1/(B_1 * D_4) * (MRR_1 - D_1**2 - D_2*sigma_R - D_3)
        sigma_E = 1/(A_1 * D_1) * (MRR_1 - D_2*sigma_R - D_3 - B_1*D_4*sigma_H)

        self.parameters = {}
        self.parameters['D_1'] = D_1
        self.parameters['D_2'] = D_2
        self.parameters['D_3'] = D_3
        self.parameters['D_4'] = D_4
        self.parameters['sigma_R'] = sigma_R
        self.parameters['sigma_H'] = sigma_H
        self.parameters['sigma_E'] = sigma_E

    def _get_Qc(self):
        '''Define PDF correction factor Qc[1]; 
        '''
        alpha_1 = self.spectral_data.alpha1
        alpha_2 = self.spectral_data.alpha2

        # Correction factor Qc is validated under following conditions
        if not 0 <= alpha_1 - alpha_2 <= 1 and 0 <= alpha_2 <= 1 and np.sqrt(1-alpha_1**2) > 0.3:
            warnings.warn('Correction factor Qc is not validated for given alpha_1 and alpha_2. Results should be evaluated carefully.')

        delta_alpha = alpha_1 - alpha_2
        Qc = 0.903 -0.28*delta_alpha + 4.448*delta_alpha**2 - 15.739*delta_alpha**3 + 19.57*delta_alpha**4 \
            -8.054*delta_alpha**5 + 1.013*alpha_2 - 4.178*alpha_2**2 + 8.362*alpha_2**3 - 7.993*alpha_2**4 + 2.886*alpha_2**5

        return Qc


import numpy as np
from scipy.integrate import quad
from .narrowband import Narrowband

class TovoBenasciutti(Narrowband):
    """Class for fatigue life estimation using frequency domain 
    method by Tovo and Benasciutti[1, 2, 3].
      
    References
    ----------
    [1] Roberto Tovo. Cycle distribution and fatigue damage under broadband
        random loading. International Journal of Fatigue, 24(11):1137{
        1147, 2002
    [2] Denis Benasciutti and Roberto Tovo. Spectral methods for lifetime
        prediction under wide-band stationary random processes. International
        Journal of Fatigue, 27(8):867{877, 2005
    [3] Denis Benasciutti and Roberto Tovo. Comparison of spectral methods for fatigue 
        analysis of broad-band Gaussian random processes. Probabilistic Engineering Mechanics,
        21(4), 287-299, 2006
    [4] Aleš Zorman and Janko Slavič and Miha Boltežar. 
        Vibration fatigue by spectral methods—A review with open-source support, 
        Mechanical Systems and Signal Processing, 2023, 
        https://doi.org/10.1016/j.ymssp.2023.110149

    Example
    -------
    Import modules, define time- and frequency-domain data

    >>> import FLife
    >>> import pyExSi as es
    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> # time-domain data
    >>> N = 2 ** 16  # number of data points of time signal
    >>> fs = 2048  # sampling frequency [Hz]
    >>> t = np.arange(0, N) / fs  # time vector
    >>> # frequency-domain data
    >>> M = N // 2 + 1  # number of data points of frequency vector
    >>> freq = np.arange(0, M, 1) * fs / N  # frequency vector
    >>> PSD_lower = es.get_psd(freq, 20, 60, variance = 5)  # lower mode of random process
    >>> PSD_higher = es.get_psd(freq, 100, 120, variance = 2)  # higher mode of random process
    >>> PSD = PSD_lower + PSD_higher # bimodal one-sided flat-shaped PSD

    Get Gaussian stationary signal, instantiate SpectralData object and plot PSD

    >>> rg = np.random.default_rng(123) # random generator seed
    >>> x = es.random_gaussian(N, PSD, fs, rg) # Gaussian stationary signal
    >>> sd = FLife.SpectralData(input=x, dt=1/fs) # SpectralData instance
    >>> plt.plot(sd.psd[:,0], sd.psd[:,1]) 
    >>> plt.xlabel('Frequency [Hz]')
    >>> plt.ylabel('PSD')

    Define S-N curve parameters and get fatigue-life estimatate

    >>> C = 1.8e+22  # S-N curve intercept [MPa**k]
    >>> k = 7.3 # S-N curve inverse slope [/]
    >>> tb = FLife.TovoBenasciutti(sd)
    >>> print(f'Fatigue life, method 1: {tb.get_life(C,k, method="method 1"):.3e} s.')
    >>> print(f'Fatigue life, method 2: {tb.get_life(C,k, method="method 2"):.3e} s.')

    Define stress vector and depict stress peak PDF

    >>> s = np.arange(0,np.max(x),.01) 
    >>> plt.plot(s,tb.get_PDF(s, method='method 1'), lw=5, alpha=.5, label = 'method 1')
    >>> plt.plot(s,tb.get_PDF(s, method='method 2'), '--', label = 'method 2')
    >>> plt.xlabel('Stress [MPa]')
    >>> plt.ylabel('PDF')
    >>> plt.legend()
    """
    def __init__(self, spectral_data):
        """Get needed values from reference object.

        :param spectral_data:  Instance of class SpectralData
        """     
        Narrowband.__init__(self, spectral_data)

    def _calculate_coefficient(self, method='method 2'):
        """Calculate weigthing parameter b for the Tovo-Benasciutti method. Parameter b is 
            defined by Tovo and Benasciutti [1,2,3].
        
        :param method:  string
            - 'method 1': `b` weighting parameter `b` is defined by Tovo[1].
            - 'method 2': `b` weighting parameter `b` is defined by Tovo and Benasciutti [2].
                          (This is the 2005 improved method [2])
            - 'method 3': `b` weighting parameter `b` is defined by Tovo and Benasciutti [3].
                          (This is the 2006 improved method [3])
        :return b: float
        """
        if method == 'method 1': 
            b = self._calculate_coefficient_method_1()
        elif method == 'method 2': 
            b = self._calculate_coefficient_method_2()
        elif method == 'method 3': 
            b = self._calculate_coefficient_method_3()
        else: 
            raise Exception('Unrecognized Input Error')
        return b


    def _calculate_coefficient_method_1(self):
        """Calculate weigthing parameter b Tovo-Benasciutti method. Parameter b is 
            defined by Tovo[1].
        
        :return b: float
        """
        alpha1 = self.spectral_data.alpha1
        alpha2 = self.spectral_data.alpha2

        b = min( (alpha1-alpha2) / (1.0 - alpha1), 1.0)
        
        return b

    def _calculate_coefficient_method_2(self):
        """Calculate weigthing parameter b for the 2005 improved Tovo-Benasciutti method. Parameter b is 
            defined by Tovo and Benasciutti [2].
        
        :return b: float
        """
        alpha1 = self.spectral_data.alpha1
        alpha2 = self.spectral_data.alpha2

        b = (alpha1-alpha2) * (  1.112 * ( 1+ alpha1*alpha2 - (alpha1+alpha2)  ) * np.exp(2.11*alpha2) +(alpha1-alpha2) ) / ((alpha2-1)**2)

        return b
        
    def _calculate_coefficient_method_3(self):
        """Calculate weigthing parameter b for the 2006 improved Tovo-Benasciutti method. Parameter b is 
            defined by Tovo and Benasciutti [3].
        
        :return b: float
        """
        alpha075 = self.spectral_data.alpha075
        alpha2 = self.spectral_data.alpha2

        b = (alpha075**2-alpha2**2) / (1-alpha2**2)

        return b

    def get_PDF(self, s, method='method 2'):
        """Returns cycle PDF(Probability Density Function) as a function of stress s.

        :param s:  numpy.ndarray
            Stress vector.
        :param method:  string

            - 'method 1': weighting parameter `b` is defined by Tovo[1].
              
            - 'method 2': weighting parameter `b` is defined by Tovo and Benasciutti [2].

            - 'method 3': weighting parameter `b` is defined by Tovo and Benasciutti [3].

        :return: function pdf(s)
        """
        alpha2 = self.spectral_data.alpha2
        m0 = self.spectral_data.moments[0]

        b = self._calculate_coefficient(method=method)

        def pdf(s):
            px = b * ((s / m0) * np.exp( - s**2 / (2 * m0))) + \
                (1 - b) * ((s / (m0 * alpha2**2)) * np.exp( - s**2 / (2 * alpha2**2 * m0))) 
            return px
        return pdf(s)

    def get_life(self, C, k, method='method 2', integrate_pdf=False):
        """Calculate fatigue life with parameters C, k, as defined in [4].

        :param C: [int,float]
            S-N curve intercept [MPa**k].
        :param k: [int,float]
            S-N curve inverse slope [/].
        :param method:  string
        
            - 'method 1': weighting parameter `b` is defined by Tovo[1].
            - 'method 2': weighting parameter `b` is defined by Tovo and Benasciutti [2].
            - 'method 3': weighting parameter `b` is defined by Tovo and Benasciutti [3].

        :param integrate_pdf:  boolean
            If true the the fatigue life is estimated by integrating the PDF, 
            Default is false which means that the theoretical equation is used
        :return:
            Estimated fatigue life in seconds.
        :rtype: float
        """ 
        if integrate_pdf:
            d = self.spectral_data.nu / C * \
                quad(lambda s: s**k*self.get_PDF(s, method=method), 
                     a=0, b=np.inf)[0]
        else:
            m0 = self.spectral_data.moments[0]
            nu = self.spectral_data.nu
            alpha2 = self.spectral_data.alpha2
            
            b = self._calculate_coefficient(method=method)

            dNB = self.damage_intesity_NB(m0=m0, nu=nu, C=C, k=k) 
            l = b + ( 1.0 - b ) * alpha2**(k-1.0)
            d = dNB * l
        
        T = 1.0/d
        return T


import numpy as np
from scipy import special
from scipy.integrate import quad

class Dirlik(object):
    """Class for fatigue life estimation using frequency domain method by Dirlik [1].
   
    References
    ----------
    [1] Turan Dirlik. Application of computers in fatigue analysis. PhD thesis,
        University of Warwick, 1985
    [2] Aleš Zorman and Janko Slavič and Miha Boltežar. 
        Vibration fatigue by spectral methods—A review with open-source support, 
        Mechanical Systems and Signal Processing, 2023, 
        https://doi.org/10.1016/j.ymssp.2023.110149
        
    Example
    -------
    Import modules, define time- and frequency-domain data

    >>> import FLife
    >>> import pyExSi as es
    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> # time-domain data
    >>> N = 2 ** 16  # number of data points of time signal
    >>> fs = 2048  # sampling frequency [Hz]
    >>> t = np.arange(0, N) / fs  # time vector
    >>> # frequency-domain data
    >>> M = N // 2 + 1  # number of data points of frequency vector
    >>> freq = np.arange(0, M, 1) * fs / N  # frequency vector
    >>> PSD_lower = es.get_psd(freq, 20, 60, variance = 5)  # lower mode of random process
    >>> PSD_higher = es.get_psd(freq, 100, 120, variance = 2)  # higher mode of random process
    >>> PSD = PSD_lower + PSD_higher # bimodal one-sided flat-shaped PSD

    Get Gaussian stationary signal, instantiate SpectralData object and plot PSD

    >>> rg = np.random.default_rng(123) # random generator seed
    >>> x = es.random_gaussian(N, PSD, fs, rg) # Gaussian stationary signal
    >>> sd = FLife.SpectralData(input=x, dt=1/fs) # SpectralData instance
    >>> plt.plot(sd.psd[:,0], sd.psd[:,1]) 
    >>> plt.xlabel('Frequency [Hz]')
    >>> plt.ylabel('PSD')

    Define S-N curve parameters and get fatigue-life estimatate

    >>> C = 1.8e+22  # S-N curve intercept [MPa**k]
    >>> k = 7.3 # S-N curve inverse slope [/]
    >>> dirlik = FLife.Dirlik(sd) # Dirlik's fatigue-life estimator
    >>> print(f'Dirlik: {dirlik.get_life(C,k):.3e} s.')   

    Define stress vector and depict peak stress PDF

    >>> s = np.arange(0,np.max(x),.01)
    >>> plt.plot(s,dirlik.get_PDF(s))
    >>> plt.xlabel('Stress [MPa]')
    >>> plt.ylabel('PDF')
    """
    def __init__(self, spectral_data):
        """Get needed values from reference object.

        :param spectral_data:  Instance of class SpectralData
        """          
        self.spectral_data = spectral_data
        self._calculate_coefficients() 

    def _calculate_coefficients(self):
        """Calculate coefficients for Dirlik method [2].
        """
        m0,m1,m2, _, m4 = self.spectral_data.moments
        
        c = np.empty(8)
        c[0] = ( 1. / np.sqrt(m0) ) 
        c[1] = ( ( m1 / m0 ) * np.sqrt( m2 / m4 ) ) 
        c[2] = ( np.sqrt( 1. / ( m0 * m4 ) ) * m2 ) 
        c[3] = ( 2. * ( c[1] - c[2]**2 ) / ( 1. + c[2]**2 ) ) 
        c[4] = ( ( c[2] - c[1] - c[3]**2 ) / (1 - c[2] - c[3] + c[3]**2 ) ) 
        c[5] = ( ( 1. - c[2] - c[3] + c[3]**2 ) / ( 1. - c[4] ) ) 
        c[6] = ( ( 1. - c[3] - c[5] ) ) 
        c[7] = ( 1.25 * ( c[2] - c[6] - c[5] * c[4] ) / c[3] ) 
        
        self.coeff = c
            
    def get_PDF(self,s):
        """Returns cycle PDF(Probability Density Function) as a function of stress s.

        :param s:  numpy.ndarray
            Stress vector
        :return: function pdf(s)
        """
        m0 = self.spectral_data.moments[0]
        Z1 = self.coeff[0]
        R = self.coeff[4]
        Q = self.coeff[7]
        G1 = self.coeff[3]
        G2 = self.coeff[5]
        G3 = self.coeff[6]

        def pdf(s):
            Z =  Z1*s
            px = (1.0/np.sqrt(m0)) * ( \
                            (G1/Q)*np.exp(-Z/Q) + \
                            ((G2*Z)/(R**2))*np.exp(-((Z)**2)/(2.*R**2)) + \
                            (G3*Z)*np.exp(-((Z)**2)/2.) \
                            )
            return px
        return pdf(s)

    def get_life(self, C, k, integrate_pdf=False):
        """Calculate fatigue life with parameters C, k, as defined in [2].

        :param C: [int,float]
            S-N curve intercept [MPa**k].
        :param k: [int,float]
            S-N curve inverse slope [/].
        :param integrate_pdf:  boolean
            If true the the fatigue life is estimated by integrating the PDF. Default is false which means that the theoretical equation is used. 
        :return: 
            Estimated fatigue life in seconds.
        :rtype: float
        """ 
        if integrate_pdf:
            d = self.spectral_data.m_p / C * \
                quad(lambda s: s**k*self.get_PDF(s), 
                     a=0, b=np.inf)[0]
        else:
            m0 = self.spectral_data.moments[0]
            
            R = self.coeff[4]
            Q = self.coeff[7]
            G1 = self.coeff[3]
            G2 = self.coeff[5]
            G3 = self.coeff[6]
            
            d = 1 / C * ( self.spectral_data.m_p 
                          * np.sqrt(m0)**k 
                          * (
                                G1 * (Q**k)*special.gamma(1.0+k)+(np.sqrt(2.0)**k)*\
                                special.gamma(1.+k/2.)*(G2 * abs(R)**k+G3)\
                            )
                        )
    
        T = 1/d
        return T


import numpy as np
from scipy.special import gamma
from scipy.integrate import quad

class Narrowband(object):
    """Class for fatigue life estimation using frequency domain 
    method by Miles [1] / Bendat and Piersol [2].
   
    References
    ----------
    [1] John W. Miles. On structural fatigue under random loading. Journal
        of the Aeronautical Sciences, 21(11):753{762, 1954.
    [2] Julius S. Bendat and Allen G. Piersol. Measurement and Analysis of Random Data.
        Wiley, 1966.
    [3] Aleš Zorman and Janko Slavič and Miha Boltežar. 
        Vibration fatigue by spectral methods—A review with open-source support, 
        Mechanical Systems and Signal Processing, 2023, 
        https://doi.org/10.1016/j.ymssp.2023.110149

    Example
    -------
    Import modules, define time- and frequency-domain data

    >>> import FLife
    >>> import pyExSi as es
    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> # time-domain data
    >>> N = 2 ** 16  # number of data points of time signal
    >>> fs = 2048  # sampling frequency [Hz]
    >>> t = np.arange(0, N) / fs  # time vector
    >>> # frequency-domain data
    >>> M = N // 2 + 1  # number of data points of frequency vector
    >>> freq = np.arange(0, M, 1) * fs / N  # frequency vector
    >>> PSD = es.get_psd(freq, 20, 60, variance = 5)  # one-sided flat-shaped PSD

    Get Gaussian stationary signal, instantiate SpectralData object and plot PSD

    >>> rg = np.random.default_rng(123) # random generator seed
    >>> x = es.random_gaussian(N, PSD, fs, rg) # Gaussian stationary signal
    >>> sd = FLife.SpectralData(input=x, dt=1/fs) # SpectralData instance
    >>> plt.plot(sd.psd[:,0], sd.psd[:,1]) 
    >>> plt.xlabel('Frequency [Hz]')
    >>> plt.ylabel('PSD')

    Define S-N curve parameters and get fatigue-life estimatate

    >>> C = 1.8e+22  # S-N curve intercept [MPa**k]
    >>> k = 7.3 # S-N curve inverse slope [/]
    >>> nb = FLife.Narrowband(sd)
    >>> print(f'Fatigue life: {nb.get_life(C,k):.3e} s.')

    Define stress vector and depict stress peak PDF

    >>> s = np.arange(0,np.max(x),.01)
    >>> plt.plot(s,nb.get_PDF(s))
    >>> plt.xlabel('Stress [MPa]')
    >>> plt.ylabel('PDF')
    """
    def __init__(self, spectral_data):
        """Get needed values from reference object.
        
        :param spectral_data:  Instance of class SpectralData
        """                        
        self.spectral_data = spectral_data

    def get_PDF(self,s):
        """Returns cycle PDF(Probability Density Function) as a function of stress s.

        :param s:  numpy.ndarray
            Stress vector.
        :return: function pdf(s)
        """
        m0 = self.spectral_data.moments[0]

        def pdf(s):
            px = (s / m0) * np.exp( - s**2 / (2 * m0)) 
            return px
        return pdf(s)

    def damage_intesity_NB(self, m0, nu, C, k):
        """Calculates narrowband damage intensity with parameters m0, nu, C, k, as defined in [2].

        :param m0: [int,float]
            Zeroth spectral moment [MPa**2].
        :param nu: [int,float]
            Frequency of positive slope zero crossing [Hz].
        :param C: [int,float]
            Fatigue strength coefficient [MPa**k].
        :param k: [int,float]
            Fatigue strength exponent [/].
        :return:
            Estimated damage intensity.   
        :rtype: float
        """
        d = nu * np.sqrt(2 * m0)**k * gamma(1.0 + k/2.0) / C  
        return  d

    def get_life(self, C, k, integrate_pdf=False):
        """Calculate fatigue life with parameters C, k, as defined in [2].

        :param C: [int,float]
            S-N curve intercept [MPa**k].
        :param k: [int,float]
            S-N curve inverse slope [/].
        :param integrate_pdf:  boolean
            If true the the fatigue life is estimated by integrating the PDF, 
            Default is false which means that the theoretical equation is used
        :return:
            Estimated fatigue life in seconds.
        :rtype: float
        """ 
        if integrate_pdf:
            d = self.spectral_data.nu / C * \
                quad(lambda s: s**k*self.get_PDF(s), 
                     a=0, b=np.inf)[0]
        else:
            m0 = self.spectral_data.moments[0]
            nu = self.spectral_data.nu
            d = self.damage_intesity_NB(m0=m0, nu=nu, C=C, k=k) 

        T = 1.0/d
        return T


from .dirlik import Dirlik
from .tovo_benasciutti import TovoBenasciutti
from .zhao_baker import ZhaoBaker
from .narrowband import Narrowband
from .alpha075 import Alpha075
from .wirsching_light import WirschingLight
from .gao_moan import GaoMoan
from .fu_cebon import FuCebon
from .low import Low
from .modified_fu_cebon import ModifiedFuCebon
from .sakai_okamura import SakaiOkamura
from .single_moment import SingleMoment
from .jiao_moan import JiaoMoan
from .bands_method import BandsMethod
from .ortiz_chen import OrtizChen
from .park import Park
from .jun_park import JunPark
from .low_2014 import Low2014
from .lotsberg import Lotsberg
from .huang_moan import HuangMoan


import numpy as np
from scipy.integrate import quad
from scipy.special import gamma

class Park(object):
    """Class for fatigue life estimation using frequency domain 
    method by Park et al.[1].
      
    References
    ----------
    [1] Jun-Bum Park, Joonmo Choung and Kyung-Su Kim. A new fatigue prediction model for marine 
        structures subject to wide band stress process. Ocean Engineering, 76: 144-151, 2014
    [2] Aleš Zorman and Janko Slavič and Miha Boltežar. 
        Vibration fatigue by spectral methods—A review with open-source support, 
        Mechanical Systems and Signal Processing, 2023, 
        https://doi.org/10.1016/j.ymssp.2023.110149
        
    Example
    -------
    Import modules, define time- and frequency-domain data

    >>> import FLife
    >>> import pyExSi as es
    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> # time-domain data
    >>> N = 2 ** 16  # number of data points of time signal
    >>> fs = 2048  # sampling frequency [Hz]
    >>> t = np.arange(0, N) / fs  # time vector
    >>> # frequency-domain data
    >>> M = N // 2 + 1  # number of data points of frequency vector
    >>> freq = np.arange(0, M, 1) * fs / N  # frequency vector
    >>> PSD_lower = es.get_psd(freq, 20, 60, variance = 5)  # lower mode of random process
    >>> PSD_higher = es.get_psd(freq, 100, 120, variance = 2)  # higher mode of random process
    >>> PSD = PSD_lower + PSD_higher # bimodal one-sided flat-shaped PSD

    Get Gaussian stationary signal, instantiate SpectralData object and plot PSD

    >>> rg = np.random.default_rng(123) # random generator seed
    >>> x = es.random_gaussian(N, PSD, fs, rg) # Gaussian stationary signal
    >>> sd = FLife.SpectralData(input=x, dt=1/fs) # SpectralData instance
    >>> plt.plot(sd.psd[:,0], sd.psd[:,1]) 
    >>> plt.xlabel('Frequency [Hz]')
    >>> plt.ylabel('PSD')

    Define S-N curve parameters and get fatigue-life estimatate

    >>> C = 1.8e+22  # S-N curve intercept [MPa**k]
    >>> k = 7.3 # S-N curve inverse slope [/]
    >>> park = FLife.Park(sd)
    >>> print(f'Fatigue life: {park.get_life(C,k):.3e} s.')

    Define stress vector and depict stress peak PDF

    >>> s = np.arange(0,np.max(x),.01) 
    >>> plt.plot(s,park.get_PDF(s))
    >>> plt.xlabel('Stress [MPa]')
    >>> plt.ylabel('PDF')
    """
    def __init__(self, spectral_data):
        """Get needed values from reference object.

        :param spectral_data:  Instance of class SpectralData
        """     
        self.spectral_data = spectral_data
        self._set_distribution_parameters() #calculate distribution parameters

    def get_PDF(self, s):
        """Returns cycle PDF(Probability Density Function) as a function of stress s.

        :param s:  numpy.ndarray
            Stress vector.
        :return: function pdf(s)
        """
        m0 = self.spectral_data.moments[0]

        def park_pdf(s):
            #PDF of stress amplitude normalized by standard deviation of process
            #half-Gaussian
            gauss_pdf = lambda s: 2/(np.sqrt(2*np.pi)*self.parameters['sigma_g'])* np.exp(-s**2/(2*self.parameters['sigma_g']**2))
            #Rayleigh
            rayleigh1_pdf = lambda s: s/self.parameters['sigma_r1']**2 * np.exp(-s**2/(2*self.parameters['sigma_r1']**2))
            #Rayleigh with unit variance
            rayleigh2_pdf = lambda s: s * np.exp(-s**2/2)

            pdf_out = self.parameters['C_g']*gauss_pdf(s) + self.parameters['C_r1']*rayleigh1_pdf(s) + self.parameters['C_r2']*rayleigh2_pdf(s)
            return pdf_out

        return 1/np.sqrt(m0) * park_pdf(s/np.sqrt(m0))


    def get_life(self, C, k, integrate_pdf=False):
        """Calculate fatigue life with parameters C, k, as defined in [2].

        :param C: [int,float]
            S-N curve intercept [MPa**k].
        :param k: [int,float]
            S-N curve inverse slope [/].
        :return:
            Estimated fatigue life in seconds.
        :rtype: float
        """ 
        m0 = self.spectral_data.moments[0]
        m_p = self.spectral_data.m_p

        if integrate_pdf:
            d = m_p / C * quad(lambda s: s**k*self.get_PDF(s), a=0, b=np.inf)[0]
        else:
            m0 = self.spectral_data.moments[0]

            d = m_p / C * (np.sqrt(2*m0))**k * (self.parameters['C_r1']*self.parameters['sigma_r1']**k * gamma(1 + k/2) \
                + self.parameters['C_r2']*gamma(1+k/2) + self.parameters['C_g']/(np.sqrt(np.pi)) * self.parameters['sigma_g']**k * gamma((1+k)/2))

        T = 1.0/d
        return T


    def _set_distribution_parameters(self):
        '''Define PDF parameters; 
        '''
        #alpha are used for n-th moment of rainflow range distrubution Mrr(n)
        alpha2 = self.spectral_data.alpha2
        alpha0_95 = self.spectral_data.get_bandwidth_estimator(self.spectral_data.PSD_splitting, i=0.95)[0]
        alpha1_97 = self.spectral_data.get_bandwidth_estimator(self.spectral_data.PSD_splitting, i=1.97)[0]
        alpha0_54 = self.spectral_data.get_bandwidth_estimator(self.spectral_data.PSD_splitting, i=0.54)[0]
        alpha0_93 = self.spectral_data.get_bandwidth_estimator(self.spectral_data.PSD_splitting, i=0.93)[0]
        alpha1_95 = self.spectral_data.get_bandwidth_estimator(self.spectral_data.PSD_splitting, i=1.95)[0]

        #Mrr(n)
        M_rr_1 = alpha2
        M_rr_2 = alpha0_95*alpha1_97
        M_rr_3 = alpha0_54*alpha0_93*alpha1_95

        #distribution parameters
        sigma_r1 = alpha2
        C_r1 = (M_rr_2 - M_rr_3) / (sigma_r1**2 * (1 - sigma_r1))
        C_r2 = (-sigma_r1*M_rr_2 + M_rr_3) / (1-sigma_r1)
        C_g = 1 - C_r1 - C_r2
        V_1 = 1/np.sqrt(np.pi) * gamma(1)/gamma(1.5)
        sigma_g = 1/(V_1*C_g) * (M_rr_1 - C_r1*sigma_r1 - C_r2)

        self.parameters = {}
        self.parameters['sigma_r1'] = sigma_r1
        self.parameters['sigma_g'] = sigma_g
        self.parameters['C_r1'] = C_r1
        self.parameters['C_r2'] = C_r2
        self.parameters['C_g'] = C_g


import numpy as np
from scipy.special import gamma
from .narrowband import Narrowband

class WirschingLight(Narrowband):
    """Class for fatigue life estimation using frequency domain 
    method by Wirsching and Light [1].
   
    References
    ----------
    [1] Paul H. Wirsching and Mark C. Light. Fatigue under wide band random
        stresses. Journal of the Structural Division, 106(7):1593-1607,
        1980
    [2] Aleš Zorman and Janko Slavič and Miha Boltežar. 
        Vibration fatigue by spectral methods—A review with open-source support, 
        Mechanical Systems and Signal Processing, 2023, 
        https://doi.org/10.1016/j.ymssp.2023.110149

    Example
    -------
    Import modules, define time- and frequency-domain data

    >>> import FLife
    >>> import pyExSi as es
    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> # time-domain data
    >>> N = 2 ** 16  # number of data points of time signal
    >>> fs = 2048  # sampling frequency [Hz]
    >>> t = np.arange(0, N) / fs  # time vector
    >>> # frequency-domain data
    >>> M = N // 2 + 1  # number of data points of frequency vector
    >>> freq = np.arange(0, M, 1) * fs / N  # frequency vector
    >>> PSD_lower = es.get_psd(freq, 20, 60, variance = 5)  # lower mode of random process
    >>> PSD_higher = es.get_psd(freq, 100, 120, variance = 2)  # higher mode of random process
    >>> PSD = PSD_lower + PSD_higher # bimodal one-sided flat-shaped PSD

    Get Gaussian stationary signal, instantiate SpectralData object and plot PSD

    >>> rg = np.random.default_rng(123) # random generator seed
    >>> x = es.random_gaussian(N, PSD, fs, rg) # Gaussian stationary signal
    >>> sd = FLife.SpectralData(input=x, dt=1/fs) # SpectralData instance
    >>> plt.plot(sd.psd[:,0], sd.psd[:,1]) 
    >>> plt.xlabel('Frequency [Hz]')
    >>> plt.ylabel('PSD')

    Define S-N curve parameters and get fatigue-life estimatate

    >>> C = 1.8e+22  # S-N curve intercept [MPa**k]
    >>> k = 7.3 # S-N curve inverse slope [/]
    >>> wl = FLife.WirschingLight(sd)
    >>> print(f'Fatigue life: {wl.get_life(C,k):.3e} s.')
    """
    def __init__(self, spectral_data):
        """Get needed values from reference object.

        :param spectral_data:  Instance of class SpectralData
        """
        Narrowband.__init__(self, spectral_data)
                           
    def get_life(self, C, k):
        """Calculate fatigue life with parameters C, k, as defined in [2].

        :param C: [int,float]
            S-N curve intercept [MPa**k].
        :param k: [int,float]
            S-N curve inverse slope [/].
        :return:
            Estimated fatigue life in seconds.
        :rtype: float
        """
        m0 = self.spectral_data.moments[0]
        nu = self.spectral_data.nu
        alpha2 = self.spectral_data.alpha2

        dNB = self.damage_intesity_NB(m0=m0, nu=nu, C=C, k=k) 
        ak = 0.926 - 0.033 * k
        bk = 1.587 * k - 2.323
        eps = np.sqrt(1 - alpha2**2)

        ro = ak + ( 1 - ak ) * ( 1 - eps )**bk
        T = 1 / (ro * dNB)
        
        return T

    def get_PDF(self, s):
        raise Exception(f'Function <get_PDF> is not available for class {self.__class__.__name__:s}.')


import numpy as np
import fatpack
import rainflow

class Rainflow(object):
    """Class for fatigue life estimation using rainflow counting method [1, 2].

    References
    ----------
    [1] C. Amzallag et. al. Standardization of the rainflow counting method for
        fatigue analysis. International Journal of Fatigue, 16 (1994) 287-293

    [2] ASTM E1049-85

    [3] Janko Slavič, Matjaž Mršnik, Martin Česnik, Jaka Javh, Miha Boltežar. 
        Vibration Fatigue by Spectral Methods, From Structural Dynamics to Fatigue Damage
        – Theory and Experiments, ISBN: 9780128221907, Elsevier, 1st September 2020

    Example
    -------
    Import modules, define time- and frequency-domain data

    >>> import FLife
    >>> import pyExSi as es
    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> # time-domain data
    >>> N = 2 ** 16  # number of data points of time signal
    >>> fs = 2048  # sampling frequency [Hz]
    >>> t = np.arange(0, N) / fs  # time vector
    >>> # frequency-domain data
    >>> M = N // 2 + 1  # number of data points of frequency vector
    >>> freq = np.arange(0, M, 1) * fs / N  # frequency vector
    >>> PSD_lower = es.get_psd(freq, 20, 60, variance = 5)  # lower mode of random process
    >>> PSD_higher = es.get_psd(freq, 100, 120, variance = 2)  # higher mode of random process
    >>> PSD = PSD_lower + PSD_higher # bimodal one-sided flat-shaped PSD

    Get Gaussian stationary signal, instantiate SpectralData object and plot PSD

    >>> rg = np.random.default_rng(123) # random generator seed
    >>> x = es.random_gaussian(N, PSD, fs, rg) # Gaussian stationary signal
    >>> sd = FLife.SpectralData(input=x, dt=1/fs) # SpectralData instance
    >>> plt.plot(sd.psd[:,0], sd.psd[:,1]) 
    >>> plt.xlabel('Frequency [Hz]')
    >>> plt.ylabel('PSD')

    Define S-N curve parameters and get fatigue-life estimatate

    >>> C = 1.8e+22  # S-N curve intercept [MPa**k]
    >>> k = 7.3 # S-N curve inverse slope [/]
    >>> rf = FLife.Rainflow(sd)
    >>> print(f'Fatigue life: {rf.get_life(C,k):.3e} s.')
    """
    def __init__(self, spectral_data, **kwargs):
        """Get needed values from reference object.

        :param spectral_data:  Instance of object SpectralData
        :param rg: Instance of numpy.random._generator.Generator
            Parameter `rg` cotrols phase of generated time history, if not already exist in spectral_data.
        """          
        self.spectral_data = spectral_data

        #set time history if not exsist in spectral_data
        if not hasattr(spectral_data, 'data'):
            rg = kwargs.get('rg', None)
            
            f = spectral_data.psd[:,0]
            psd = spectral_data.psd[:,1]
            
            # se parametra T in fs
            if 'T' and 'fs' in kwargs.keys():             
                self.spectral_data._set_time_history(f=f, psd=psd, **kwargs)
            else:
                raise Exception('Time history is not set; T and fs must be specified.')
            
    def get_life(self, C, k, algorithm = 'four-point',  Su = False, range = False, 
                 nr_load_classes=512, **kwargs):
        """Calculate fatigue life with parameters C, k, as defined in [3]
        
        :param C: [int,float]
            S-N curve intercept [MPa**k].
        :param k: [int,float]
            S-N curve inverse slope [/].
        :param algorithm: str
            Cycle counting method. Options are 'three-point' and 'four-point'.
            Defaults to 'four-point'.
        :param Su: [int,float]
            Ultimate tensile strength [MPa]. If specified, Goodman equivalent stress
            is used for fatigue life estimation. Defaults to False.
        :param range: bool
            If True, ranges instead of amplitudes are used for fatigue life estimation.
            Defaults to False.
        :param nr_load_classes: int
            The number of intervals to divide the min-max range of the dataseries
            into. Used with algorithm ='four-point'. 
            Defaults to 512.
        :returns:
            Estimated fatigue life in seconds.
        :rtype: float
        """    
        t = self.spectral_data.t
        cycles = self._get_cycles(algorithm = algorithm, Su=Su, **kwargs)
        ranges = cycles[0]
        means = cycles[1]
        
        if range == True: 
            pass
        elif range == False: 
            ranges *= 0.5
        else: 
            raise Exception('Unrecognized Input Error')

        if Su: ranges = ranges/(1. - means/Su)

        try :
            counts = cycles[2]  #three-point algorithm returns half cycles
            d = np.sum(counts * ranges**k / C)
        except Exception:
            d = np.sum(ranges**k / C)

        T = t / d
        
        return T

    def _get_cycles(self, algorithm = 'four-point', nr_load_classes=512, **kwargs):
        """
        :param algorithm: str
            Cycle counting method. Options are `three-point` and `four-point`.
            Defaults to `four-point`.
        :param nr_load_classes: int
            The number of intervals to divide the min-max range of the dataseries
            into. Used with algorithm =`four-point`. 
            Defaults to 512.
        :returns (ranges, means, counts): tuple of numpy.ndarray
            For algorithm = 'four-point', tuple with only two elemensts (ranges, means) is returned.
        """
        if algorithm == 'four-point':
            ranges, means = fatpack.find_rainflow_ranges(self.spectral_data.data, k=nr_load_classes, return_means=True)
            return ranges, means

        elif algorithm == 'three-point':
            cycles = np.array(list(rainflow.extract_cycles(self.spectral_data.data)))
            ranges = cycles[:,0]
            means = cycles[:,1]
            counts = cycles[:,2]
            return ranges, means, counts

        else: 
            raise ValueError('Set `algorithm` either to `three-point` or `four-point`.')


from matplotlib.pyplot import table
import numpy as np
from scipy import stats
import tkinter as tk
from tkinter.filedialog import asksaveasfilename
from matplotlib.figure import Figure
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
import time

def relative_error(value, value_ref):
    """Return the relative error 

    :param value: [int,float]
        Value to compare
    :param value: [int,float]
        Reference value
    :return err: float
        Relative error.
    """ 
    return (value-value_ref)/value_ref

def basquin_to_sn(Sf, b, range = False):
    """Converts Basquin equation parameters Sf and b to fatigue life parameters C and k,
    as defined in [2]. Basic form of Basquin equation is used here: Sa = Sf* (2*N)**k 

    :param Sf: [int,float]
        Fatigue strength coefficient [MPa**k].
    :param b : [int,float]
        Fatigue strength exponent [/]. Represents S-N curve slope.
    :param range : bool
        False/True sets returned value C with regards to amplitude / range count, respectively.
    :return C, k: float
        C - S-N curve intercept [MPa**k], k - S-N curve inverse slope [/].
    """ 
    if not range:
        k = -1/b 
        C = 0.5*Sf**k
        return C,k

    else: 
        k = -1/b 
        C = 0.5*(2*Sf)**k
        return C,k

def pdf_rayleigh_sum(m0L, m0H):
    """Returns PDF (Probability Density Function) of sum of 2 Rayleigh distributed RV (Random Variables).

    :param m0L, moH: [int,float]
        Variance of first and second Rayleigh distributed RV. 
    :return pdf: function
    """
    m0 = m0L + m0H
    m0L_norm = m0L/m0
    m0H_norm = m0H/m0

    def pdf(x):
        pdf = 1 / np.sqrt(m0) * (m0L_norm * x/np.sqrt(m0) * np.exp(- (x/np.sqrt(m0))**2 /(2*m0L_norm)) \
            + m0H_norm * x/np.sqrt(m0) * np.exp(- (x/np.sqrt(m0))**2 /(2*m0H_norm)) \
            + np.sqrt(2*np.pi*m0L_norm*m0H_norm) * ((x/np.sqrt(m0))**2 -1) * np.exp(-(x/np.sqrt(m0))**2/2) \
            * (stats.norm.cdf(np.sqrt(m0L_norm/m0H_norm)*x/np.sqrt(m0)) + stats.norm.cdf(np.sqrt(m0H_norm/m0L_norm)*x/np.sqrt(m0)) - 1))
        return pdf
    return pdf


def random_gaussian(freq, PSD, T, fs, rg=None, random_amplitude=False, **kwargs):
    """
    Stationary Gaussian realization of random process, characterized by PSD.
    
    Random process is obtained with IFFT of amplitude spectra with random phase [1]. Area under PSD curve represents variance of random process.
    
    :param freq: array
        Frequency vector
    :param PSD: array
        One-sided power spectral density [unit^2].
    :param T: int,float
        Length of returned random process [s].
    :param fs: int,float
        Sampling frequency [Hz].
    :param rg: numpy.random._generator.Generator
        Initialized Generator object
    :param random_amplitude: Boolean
        If true, Rayleigh distributed amplitude is used in addition 
        to uniformly distributed phase. Defaults to False
    :return: t, signal
        Time and stationary Gaussian realization of random process
    
    Notes
    -----
    PSD data points are interpolated using numpy.interp() function. Option for 
    interpolation function will be added in the future.
    
    References
    ----------
    [1] D. E. Newland. An Introduction to Random Vibrations, Spectral & Wavelet Analysis.
    Dover Publications, 2005
    """ 
    # time and frequency data
    if np.__version__>='2.0.0':
        trapezoid = np.trapezoid
    else:
        trapezoid = np.trapz
    var = trapezoid(PSD,freq)
    N = int(T * fs)
    M = N//2 + 1
    t = np.arange(0,N) / fs # time vector

    M = N//2 + 1
    freq_new = np.arange(0, M, 1) / N  * fs # frequency vector
    PSD_new = np.interp(freq_new, freq, PSD, left=0, right=0)
    
    if rg == None:
        rg = np.random.default_rng()
    
    if isinstance(rg, np.random._generator.Generator):
        
        phase = rg.uniform(0, 1, len(PSD_new))
        
        ampl_spectra = np.sqrt(PSD_new * N * fs / 2)  # amplitude spectra modulus
        if random_amplitude == True:
            ampl_spectra = np.array([rg.rayleigh(size=1, scale=x)[0] for x in ampl_spectra]) # Rayleigh distributed amplitudes

        ampl_spectra_random = ampl_spectra * np.exp(
            1j * phase * 2 * np.pi)  # amplitude spectra,  random phase

    else:
        raise ValueError(
            '`rg` must be initialized Generator object (numpy.random._generator.Generator)!'
        )

    signal = np.fft.irfft(ampl_spectra_random, n=N)  # time signal
    signal = signal/np.std(signal) * var**.5
    return t, signal


class PSDgen(object):
    """
    Get Power Spectral Density (PSD) via graph or table.

    PSD is determined by linear interpolation of selected points, in linear 
    or logarithmic scale. 
    """
    def __init__(self):
        self.data_entries = {} # dictionary for point data entries
        self.scale = 'lin' # 'lin' - linear, 'log' - logarithmic
        self.deltaF_lock = False # disable change of freq resolution after firts point selection
        self.shift_is_held = False
        self.D_is_held = False
        self._button_press_data = False 
        self._data_point_drag = False
        
        # Tkinter
        self.root = tk.Tk()
        self.root.title("FLife: PSD input")
        self.root.bind_all("<1>", lambda event:event.widget.focus_set()) #Set focus on widget when cklicked, left
        self.root.bind_all("<2>", lambda event:event.widget.focus_set()) #Set focus on widget when cklicked, midddle
        self.root.bind_all("<3>", lambda event:event.widget.focus_set()) #Set focus on widget when cklicked, right

        # Tkinter layout; grid() geometry manager. Top level grid is one row and three columns
        # Grid resizing settings
        self.root.rowconfigure(0, minsize=400, weight=1) #height of first row, i.e. window height
        self.root.columnconfigure(0, minsize=80, weight=0) #width of first column
        self.root.columnconfigure(1, minsize=700, weight=1) #width of second column
        self.root.columnconfigure(2, minsize=100, weight=0) #width of third column

        # First column; menu: help, export and confirm buttons
        frm_menu = tk.Frame(self.root)
        # Help button
        btn_help = tk.Button(frm_menu, text="Help", command=self.help)
        btn_help.grid(row=0, column=0, sticky="ew", padx=5, pady=5)
        # Export PSD button
        btn_export = tk.Button(frm_menu, text="Export PSD", command=self.save_file)
        btn_export.grid(row=1, column=0, sticky="ew", padx=5, pady=5)
        # Add frame frm_menu to grid
        frm_menu.grid(row=0, column=0, sticky="nsew")

        # Second column; frame for help text, canvas, toolbar and scale
        frm_PSD_display = tk.Frame(self.root)
        # Frame resizing settings
        frm_PSD_display.grid_rowconfigure(1, weight=1) # Canvas
        frm_PSD_display.grid_columnconfigure(0, weight=1)
        # Help
        self.lbl_help = tk.Label(frm_PSD_display)
        self.lbl_help.grid(row=0, column=0, sticky='nsew', pady = 5)
        self.text_initial = 'SHIFT+LMB click to add point, SHIFT+RMB click&drag to move point, SHIFT+D+LMB click to remove data point.'
        self.lbl_help['text'] = self.text_initial
        #Figure and Canvas widget
        self.fig = Figure(figsize=(7, 4))
        self.freq_max = 10000 # [Hz]
        self.PSD_max = 100 # [Unit^2/H]
        self.dfreq = 1.0 # [Hz], default value
        self.freq = np.arange(0, self.freq_max, self.dfreq)
        self.ax = self.fig.add_subplot(111)
        self.ax.grid(True, which='both')
        self.ax.set_xlim([0,self.freq_max])
        self.ax.set_ylim([0,self.PSD_max])
  
        # Integrate matplotlib figure
        self.canvas = FigureCanvasTkAgg(self.fig, master=frm_PSD_display)  
        canvas_widget = self.canvas.get_tk_widget()
        # Add canvas to frame
        canvas_widget.grid(row=1, column=0, sticky="nsew")
      	# Connecting functions to event manager
        self.fig.canvas.mpl_connect('key_press_event', self.on_key_press)
        self.fig.canvas.mpl_connect('key_release_event', self.on_key_release)
        self.fig.canvas.mpl_connect('button_press_event', self.select_data_point)
        self.fig.canvas.mpl_connect('button_release_event', self.move_data_point)
        self.fig.canvas.mpl_connect('motion_notify_event', self.drag_data_point)

        # Toolbar  widget
        toolbar = NavigationToolbar2Tk(self.canvas, frm_PSD_display, pack_toolbar=False)
        # Add toolbar to frame
        toolbar.grid(row=2, column=0, sticky='nsw')

        #Scale buttons are contained in frame frm_scale
        frm_scale = tk.Frame(frm_PSD_display)
        btn_scale_lin = tk.Button(frm_scale, text="lin", command=self.set_scale_lin)
        btn_scale_log = tk.Button(frm_scale, text="log", command=self.set_scale_log)
        btn_scale_lin.grid(row=0, column=1, sticky="snew", padx=2, pady=7)
        btn_scale_log.grid(row=0, column=2, sticky="snew", padx=2, pady=7)
        # Add frm_scale to frm_PSD_display
        frm_scale.grid(row=2, column=0, sticky="sne")
        # Add frame frm_PSD_display to grid
        frm_PSD_display.grid(row=0, column=1, sticky="nsew")

        # Third column; frame for frequency resolution button, PSD table and Confirm PSD Button
        frm_column3 = tk.Frame(self.root)
        frm_column3.grid_rowconfigure(1, weight=1) # configure table size settings
        # Frequency resolution button frame
        # label
        frm_deltaf = tk.Frame(frm_column3)
        lbl_deltaF = tk.Label(frm_deltaf, text='df [Hz]', width=5, relief=tk.FLAT)
        lbl_deltaF.grid(row=0, column=0, sticky='nsw', pady=8)
        #Entry
        self.ent_deltaF = tk.Entry(frm_deltaf, width=10)
        self.ent_deltaF.insert(0, str(self.dfreq))
        self.ent_deltaF.grid(row=0, column=1, sticky = 'nsw', pady=8, padx=2)
        #button
        btn_deltaF = tk.Button(frm_deltaf, text='\N{RIGHTWARDS BLACK ARROW}', command=self.confirm_deltaf)
        btn_deltaF.grid(row=0, column=2, sticky='nsw', pady=8, padx=2)
        #add frm frm_deltaF to frm_column3
        frm_deltaf.grid(row=0, column=0)
        # Frame for PSD table
        self.frm_PSD_table = tk.Frame(frm_column3)
        # 0.th row: column name and units
        lbl_nmbr = tk.Label(self.frm_PSD_table, width=3, relief=tk.FLAT, text='#')
        lbl_nmbr.grid(row=0, column=0, sticky='nswe', padx=1, pady=4)
        lbl_freq = tk.Label(self.frm_PSD_table, width=5, relief=tk.FLAT, text='f [Hz]')
        lbl_freq.grid(row=0, column=1, sticky='nswe', padx=1, pady=4)
        lbl_PSD = tk.Label(self.frm_PSD_table, width=12, relief=tk.FLAT, text='PSD [Unit^2/Hz]')
        lbl_PSD.grid(row=0, column=2, sticky='nswe', padx=1, pady=4)
        # 1.st row is empty at initialization
        # data-point number
        lbl_nmbr = tk.Label(self.frm_PSD_table, width=1, relief=tk.SUNKEN)
        lbl_nmbr['text']='1.'
        lbl_nmbr.grid(row=1, column=0, sticky='nswe', padx=1, pady=1)
        #frequency
        num_freq  = tk.StringVar()
        ent_freq = tk.Entry(self.frm_PSD_table, textvariable=num_freq, width=10)
        ent_freq.grid(row=1, column=1, sticky='nswe', padx=0, pady=1)
        #PSD value
        num_PSD  = tk.StringVar()
        ent_PSD= tk.Entry(self.frm_PSD_table, textvariable=num_PSD, width=10)
        ent_PSD.grid(row=1, column=2, sticky='nswe', padx=0, pady=1)
        #save to variable
        self.new_entry =  [None, None, lbl_nmbr, ent_freq, ent_PSD] # use for new entry via table
        #confirm button
        self.enter_PSD = tk.Button(self.frm_PSD_table, text="Confirm table")
        self.enter_PSD.bind('<Button-1>', self.add_data_from_table)
        # Add confirm buttor to frm_PSD_table
        self.enter_PSD.grid(row=2, column=2, sticky='nse', padx=1, pady=1)
        # Add frame frm_PSD_table to frm_column3
        self.frm_PSD_table.grid(row=1, column=0, sticky='news')
        # Confirm PSD button
        btn_confirm = tk.Button(frm_column3, height = 1, text="Confirm PSD and exit", command=self.on_closing)
        btn_confirm.grid(row=2, column=0, sticky="nsew", padx=5, pady=7)
        # Add frm_column3 to grid
        frm_column3.grid(row=0, column=2, sticky='nsew', )

        self.root.protocol("WM_DELETE_WINDOW", self.on_closing)
        self.root.mainloop()

    def help(self):
        window = tk.Toplevel(self.root)

        # Title
        window.title("FLife: PSD input - Help")
        lbl_help = tk.Label(window, width=70, height=10)
        help_text = [
            " Set frequency resolution df with click on right arrow button before selecting data points.\
            \n\nAdd a point with SHIFT + LEFT mouse button.\
            \nTranslate a point with SHIFT + RIGHT mouse button hold.\
            \nDelete a point with SHIFT + D + LEFT mouse button.\
            \n\nPoint can be entered (or changed) via PSD table with click on `Confirm table'.\
            \n\nData point selection is activated by mouse button click on graph area."
        ]
        lbl_help['text'] = ''.join(help_text)
        lbl_help.pack(padx=10, pady=10)

    def save_file(self):
        """
        Save the current PSD as a new file.
        """
        filepath = asksaveasfilename(
            defaultextension="txt",
            filetypes=[("Text Files", "*.txt"), ("All Files", "*.*")],
        )
        if not filepath:
            return
        # If data points exist
        try:
            PSD = self.get_PSD()
            # Save to txt
            np.savetxt(filepath, PSD, fmt=['%d','%d'])
        except:
            pass
        self.root.title(f"FLife: PSD input - {filepath}")

    def confirm_deltaf(self):
        """
        Set frequency resolution for PSD. Defaults to 1 Hz.
        """
        #unlock change in freq resolution
        if len(self.data_entries) == 0:
            self.dfreq = float(self.ent_deltaF.get())
            self.ent_deltaF.delete(0, tk.END)
            self.ent_deltaF.insert(0, f'{self.dfreq:.1f}')
            self.freq = np.arange(0, self.freq_max, self.dfreq)
        else:
            self.ent_deltaF.config(background="RED")
            self.root.update()
            time.sleep(0.1)
            self.ent_deltaF.delete(0, tk.END)
            self.ent_deltaF.insert(0, f'{self.dfreq:.1f}')
            self.ent_deltaF.config(background="WHITE")            

    def get_PSD(self):
        """
        Return PSD, defined by linear interpolation of entered data points.
        """
        # If data points exist
        try:
            # Unpack data points
            list_freq, list_psd, list_lbl_nmbr, list_ent_freq, list_ent_PSD = self._list_unpack_data_entries()
            freq_data = np.array(list_freq)
            psd_data =  np.array(list_psd)
            # Interpolate data to frequency vector, from first to last frequency in data points
            freq_min_index = np.where(self.freq == freq_data[0])[0][0]
            freq_max_index = np.where(self.freq == freq_data[-1])[0][0]
            psd_values = np.interp(self.freq[freq_min_index:freq_max_index], freq_data, psd_data)
            PSD = np.column_stack([self.freq[freq_min_index:freq_max_index], psd_values])
        except:
            # Return zero valued PSD
            psd_values = np.zeros_like(self.freq)
            PSD = np.column_stack([self.freq, psd_values])
        return PSD

    def on_closing(self):
        self.root.destroy()

    def on_key_press(self, event):
        """
        Function triggered on key press (SHIFT, D).
        """
        # Change flag
        if event.key == 'shift':
            self.shift_is_held = True
        elif event.key == 'd':
            self.D_is_held = True
        elif event.key == 'D':
            self.D_is_held = True
        self.interactive_help()

    def on_key_release(self, event):
        """
        Function triggered on key release (SHIFT, D).
        """
        if event.key == 'shift':
            self.shift_is_held = False
        elif event.key == 'd':
            self.D_is_held = False
        elif event.key == 'D':  
            self.D_is_held = False
        self.interactive_help()

    def interactive_help(self):
        # Change lbl_help text
        if self.shift_is_held == True and self.D_is_held == False:
            self.lbl_help['text'] = 'LMB click to add data point, RMB click&drag to move data point, D+LMB click to remove data point.'
        elif self.shift_is_held == True and self.D_is_held == True:
            self.lbl_help['text'] = 'LMB click to delete data point.'
        elif self.shift_is_held == False and self.D_is_held == True:
            self.lbl_help['text'] = 'SHIFT+LMB click to remove data point.'
        elif self.shift_is_held == False and self.D_is_held == False:
            self.lbl_help['text'] = self.text_initial

    def set_scale_lin(self):
        self.scale= 'lin'
        self.plot_PSD()
        
    def set_scale_log(self):
        self.scale= 'log'
        self.plot_PSD()
    
    def plot_PSD(self):
        """
        Plot PSD points, linearly interpolated.
        """
        # get current axes limit
        x_lim = self.ax.get_xlim()
        y_lim = self.ax.get_ylim()
            
        try: # if at least one data point exist, unpack data_entries
            list_freq, list_psd, list_lbl_nmbr, list_ent_freq, list_ent_PSD = self._list_unpack_data_entries() 
        except: # no data
            list_freq, list_psd = [], []

        # Plot
        self.ax.clear()
        if self.scale == 'lin':
            self.ax.plot(list_freq, list_psd, 'ko-', ms=3)
        elif self.scale == 'log':
            self.ax.loglog(list_freq, list_psd, 'ko-', ms=3)

        self.ax.grid(True, which='both')
        self.ax.set_xlim(x_lim)
        self.ax.set_ylim(y_lim)
        self.fig.canvas.draw()

    def _unpack_data_entries(self):
        """
        Sorts and unpack self.data_entries. Returned data are sorted by increasing frequency.
        """
        sorted_input = sorted(self.data_entries.items(), key=lambda item: item[1][0]) #sorted freqeuncy; data_entries.values[0]
        tuples = zip(*sorted_input)
        data_entries_nmbr, data_entries = [list(tuple) for tuple in  tuples] #split consecutive input number and data
        return data_entries

    def _list_unpack_data_entries(self):
        """
        Returneds list of data, sorted by increasing frequency.
        """
        data_entries_data = self._unpack_data_entries()
        list_data_entries = [list(data) for data in zip(*data_entries_data)]
        return list_data_entries
        
    def select_data_point(self, event):
        """
        Select data point (freqeuncy, value) from graph.
        """
        try:  #if mouse is over graph
            y_data = float(event.ydata)
            x_data = float(event.xdata)

            if event.button == 1 and self.shift_is_held and not self.D_is_held:
                self.add_data_from_graph(x_data,y_data)
                self.refresh_PSD_table()
                self.plot_PSD()
            elif event.button == 1 and self.shift_is_held and self.D_is_held:
                self.remove_data(x_data)
                self.refresh_PSD_table()
                self.plot_PSD()
            elif event.button == 3 and self.shift_is_held:
                self._button_press_data = event  # save data point to be translated
        except:
            pass

    def move_data_point(self, event):
        """
        Moves selected data point to another location on graph. Frequency and PSD value are updated, accordingly.
        Function is triggered by release of RIGHT mouse button.
        """
        try:  #if mouse is over Figure object
            y_data = float(event.ydata)
            x_data = float(event.xdata)
            
            if event.button == 3 and self.shift_is_held and self._button_press_data:
                self._drag_data_point(x_data, y_data, flag = False) #False flag disables moving_data_point() 
                self._button_press_data = False
                # replot PSD
                self.plot_PSD()
                # refresh Table
                self.refresh_PSD_table()
        except:
            pass

    def _drag_event_handle(self, event):
        """
        Inner function for use with drag_data_point(). When drag_data_point() is running, motion_notify_event is funneled to 
        this function.
        """
        if self._drag_data_point == False:
            #self.root.update()
            # rebind event back to func drag_data_point
            self.fig.canvas.mpl_connect('motion_notify_event', self.drag_data_point)
        else:
            pass

    def drag_data_point(self, event):
        """
        Updates data point while mouse is moving and RIGHT mouse button is held.
        """
        if self._data_point_drag == True:
            # rebind event to func _drag_event_handle when drag event is in progress
            self.fig.canvas.mpl_connect('motion_notify_event', self._drag_event_handle)
            pass
        else:
            try:  #if mouse is over Figure object
                y_data = float(event.ydata)
                x_data = float(event.xdata)

                if event.button == 3 and self.shift_is_held and self._button_press_data:
                    self._data_point_drag = True
                    self._drag_data_point(x_data, y_data, flag = event)
                    self.root.update()
                    # set flag
                    self._data_point_drag = False
                    # replot PSD
                    self.plot_PSD()
                    # refresh Table
                    self.refresh_PSD_table()
            except:
                pass

    def _drag_data_point(self, x_data, y_data, flag):
        """
        Subfunction used in functions move_data_point and drag_data_point. 
        """
        x_last = self._button_press_data.xdata

        # Unpack data_entries
        list_freq, list_psd, list_lbl_nmbr, list_ent_freq, list_ent_PSD = self._list_unpack_data_entries()

        # Get old frequency 
        x_ind = np.abs(x_last - self.freq).argmin()
        freq_old = self.freq[x_ind]

        #new freqeuncy
        x_ind = np.abs(x_data - self.freq).argmin()
        freq_new = self.freq[x_ind]

        if freq_old in list_freq:
            indx = list_freq.index(freq_old)
            list_freq[indx] = freq_new
            list_psd[indx] = y_data

            data_entries = dict([(indx,u) for indx,u in enumerate(zip(list_freq, list_psd, list_lbl_nmbr, list_ent_freq, list_ent_PSD))])
            self.data_entries = dict(sorted(data_entries.items(), key=lambda item: item[1][0])) #sorted freqeuncy; data_entries.values[0]

        self._button_press_data = flag

    def remove_data(self, x_data):
        # Get selected frequency 
        list_freq, list_psd, list_lbl_nmbr, list_ent_freq, list_ent_PSD = self._list_unpack_data_entries()
        freq_sel = self._data_select_zoom(x_data, list_freq, 0.05)

        for_deletion = None
        for key in self.data_entries.keys(): 
            if freq_sel == self.data_entries[key][0]:
                for_deletion = key
                
        if for_deletion !=  None:
            deleted = self.data_entries.pop(for_deletion)
            # remove data point from table
            lbl_nmbr = deleted[2]
            lbl_nmbr.grid_remove()
            ent_freq = deleted[3]    
            ent_freq.grid_remove() 
            ent_PSD = deleted[4]    
            ent_PSD.grid_remove()

    def _data_select_zoom(self, x_data, list_freq, zoom):
            """
            Select closest frequency.
            """
            freq_sel_index = np.abs(x_data - self.freq).argmin()
            freq_sel = self.freq[freq_sel_index]
            index_min, index_max = self.ax.get_xlim()

            # Frequency index range
            delta_index = index_max - index_min
            zoom_range = int(zoom * delta_index)
            if zoom_range%2 == 0:
                pass
            else:
                zoom_range += 1
            freq_zoom_indexes = np.arange(freq_sel_index-int(zoom_range/2), freq_sel_index+int(zoom_range/2)+1, 1)
            # 
            zoomed_frequencies = []
            for freq in list_freq:
                if freq in self.freq[freq_zoom_indexes]:
                    delta = np.abs(freq_sel - freq)
                    zoomed_frequencies.append((freq, delta))
            #sort to closest frequency
            zoomed_frequencies = sorted(zoomed_frequencies, key=lambda x: x[1])

            try:
                freq = zoomed_frequencies[0][0]
            except:
                freq = freq_sel
            return freq

    def _sort_data_entries(self):
        """
        Function sort and updates data entries. Sorting is done according to frequency, keys are updated.
        """
        list_freq, list_psd, list_lbl_nmbr, list_ent_freq, list_ent_PSD = self._list_unpack_data_entries()
        # update keys, in case of deleted data point
        data_entries = dict([(indx,u) for indx,u in enumerate(zip(list_freq, list_psd, list_lbl_nmbr, list_ent_freq, list_ent_PSD))])
        self.data_entries = dict(sorted(data_entries.items(), key=lambda item: item[1][0])) #sorted frequency; data_entries.values[0]

    def add_data_from_table(self, event):
        """
        Add data from table to self.data_entries
        """
        # Read table and update data entries
        table_len = len(self.data_entries) 
        # Sort data
        if table_len > 0:
            self._sort_data_entries()

        index = 0 #if table_len = 0, index=0
        for index in range(table_len): 
            freq, PSD, lbl_nmb, ent_freq, ent_PSD = self.data_entries[index]
            freq = float(ent_freq.get())
            PSD = float(ent_PSD.get())
            self.data_entries[index] = [freq, PSD, lbl_nmb, ent_freq, ent_PSD] 
        
        try:
            if float(self.new_entry[3].get()) >= 0 and float(self.new_entry[4].get()) >= 0: # if new data point is given
                # read entered data point and add to data entries
                new_ent_freq = self.new_entry[3]
                new_ent_PSD = self.new_entry[4]
                freq_new = float(new_ent_freq.get())
                PSD_new = float(new_ent_PSD.get())
                self._add_data(freq_new, PSD_new)
                # Clear new entry text
                new_ent_freq.delete(0, 'end')
                new_ent_PSD.delete(0, 'end')
        except:
            pass

        # Refresh Table
        self.refresh_PSD_table()
        # Replot PSD
        self.plot_PSD()
        
    def add_data_from_graph(self, x_data, y_data):
        """
        Adds data from clik on graph.
        """
        self._add_data(x_data, y_data)

    def _add_data(self, freq_input, psd_input):
        """
        Read input data and adds to dict self.data_entries
        """
        try: # if at least one data point exist, unpack data_entries
            list_freq, list_psd, list_lbl_nmbr, list_ent_freq, list_ent_PSD = self._list_unpack_data_entries() 
        except: # else create empyt lists
            list_freq, list_psd, list_lbl_nmbr, list_ent_freq, list_ent_PSD = [list() for u in range(5)] 

        # Get selected frequency 
        # if bigger then max freq
        if freq_input > self.freq[-1]:
            self.freq = np.arange(0, freq_input + self.dfreq, self.dfreq)
            self.freq_max = self.freq[-1]

        x_ind = np.abs(freq_input - self.freq).argmin()
        freq_sel = self.freq[x_ind]
        
        # Check axes limit
        if freq_sel >= self.freq_max*0.95:
            self.freq_max = freq_sel * 1.1
            self.freq = np.arange(0, self.freq_max, self.dfreq)
        if psd_input >= self.PSD_max*0.95:
            self.PSD_max = psd_input*1.1

        # update PSD_data; if freq_sel already in PSD-data, replace, otherwise append
        if freq_sel in list_freq:
            pass # if data point with freq_sel already exist, it should be dragged to another point on graph
        else:
            list_freq.append(freq_sel) 
            list_psd.append(psd_input)
            # Add widgets
            row_indx = len(list_freq) + 1
            # data-point number
            lbl_nmbr = tk.Label(self.frm_PSD_table, width=1, relief=tk.SUNKEN)
            lbl_nmbr['text']= str(row_indx)
            lbl_nmbr.grid(row=row_indx, column=0, sticky='nswe', padx=1, pady=1)
            list_lbl_nmbr.append(lbl_nmbr)
            # frequency
            num_freq  = tk.StringVar()
            ent_freq = tk.Entry(self.frm_PSD_table, textvariable=num_freq, width=10)
            ent_freq.grid(row=row_indx, column=1, sticky='nswe', padx=0, pady=1)
            list_ent_freq.append(ent_freq)
            # PSD value
            num_PSD  = tk.StringVar()
            ent_PSD= tk.Entry(self.frm_PSD_table, textvariable=num_PSD, width=10)
            ent_PSD.grid(row=row_indx, column=2, sticky='nswe', padx=0, pady=1)
            list_ent_PSD.append(ent_PSD)
            # confirm button
            self.enter_PSD.grid(row=row_indx+1, column=2, sticky='nse', padx=1, pady=1)
            # sort and save
            data_entries = dict([(indx,u) for indx,u in enumerate(zip(list_freq, list_psd, list_lbl_nmbr, list_ent_freq, list_ent_PSD))])
            self.data_entries = dict(sorted(data_entries.items(), key=lambda item: item[1][0])) # sorted frequency; data_entries.values[0]

    def refresh_PSD_table(self):
        """Updates PSD table, based on self.data_entries
        """
        try:
            # Unpack data_entries
            data_entries = self._unpack_data_entries()

            #update widgets
            for row_indx, entry_data in enumerate(data_entries):
                row_indx += 1 #zero-th row is used for column description
                #freq and PSD value
                freq = f'{entry_data[0]:.0f}'
                PSD = f'{entry_data[1]:.1f}'
                #Input point number
                lbl_nmbr = entry_data[2]
                lbl_nmbr['text']= str(row_indx)
                lbl_nmbr.grid(row=row_indx, column=0, sticky='nswe', padx=1, pady=1)
                #Freq
                ent_freq = entry_data[3]
                ent_freq.delete(0, tk.END)
                ent_freq.insert(0, freq)
                ent_freq.grid(row=row_indx, column=1, sticky='nswe', padx=0, pady=1)
                #PSD
                ent_PSD = entry_data[4]
                ent_PSD.delete(0, tk.END)
                ent_PSD.insert(0, PSD)
                ent_PSD.grid(row=row_indx, column=2, sticky='nswe', padx=0, pady=1)
        
            #Change row of entry line
            self.new_entry[2]['text'] = str(row_indx+1)
            self.new_entry[2].grid(row=row_indx+1, column=0, sticky='nswe', padx=0, pady=1)
            self.new_entry[3].grid(row=row_indx+1, column=1, sticky='nswe', padx=0, pady=1)
            self.new_entry[4].grid(row=row_indx+1, column=2, sticky='nswe', padx=0, pady=1)
        
        except:
            self.new_entry[2]['text'] = str(1)
            self.new_entry[2].grid(row=1, column=0, sticky='nswe', padx=0, pady=1)
            self.new_entry[3].grid(row=1, column=1, sticky='nswe', padx=0, pady=1)
            self.new_entry[4].grid(row=1, column=2, sticky='nswe', padx=0, pady=1)
